# -*- coding: utf-8 -*-
"""
lories.application.view.pages.page
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
import re
from abc import ABC, ABCMeta, abstractmethod
from functools import wraps
from typing import Any, Optional

import dash

import pandas as pd
from lories.application import InterfaceException
from lories.application.view.pages.layout import PageLayout
from lories.util import validate_key


class PageMeta(ABCMeta):
    def __call__(cls, *args, **kwargs):
        page = super().__call__(*args, **kwargs)
        cls._wrap_method(page, "create_layout")
        cls._wrap_method(page, "register")

        return page

    # noinspection PyShadowingBuiltins
    @staticmethod
    def _wrap_method(object: Any, method: str) -> None:
        _wrap_method = getattr(object, f"_do_{method}")
        _run_method = getattr(object, method)
        setattr(object, f"_run_{method}", _run_method)
        setattr(object, method, _wrap_method)


# noinspection PyShadowingBuiltins
class Page(ABC, metaclass=PageMeta):
    id: str
    key: str
    name: str
    title: str
    description: Optional[str]

    order: int = 100

    group: Optional[Page]

    _layout: PageLayout
    _created: bool = False
    _registered: bool = False

    def __init__(
        self,
        id: str,
        key: str,
        name: str,
        title: Optional[str] = None,
        description: Optional[str] = None,
        order: Optional[int] = None,
        group: Optional[Page] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__()
        self._logger = logging.getLogger(self.__module__)

        self.id = self._encode_id(id)
        self.key = validate_key(key)
        self.name = name
        self.title = title if title is not None else name
        self.description = description

        self.order = Page.order if pd.isna(order) else order
        self.group = group
        self._layout = PageLayout(id=f"{self.id}-container")

    @property
    def path(self) -> str:
        _path = f"/{self._encode_id(self.key)}"
        return _path if self.group is None else self.group.path + _path

    @property
    def layout(self) -> PageLayout:
        return self._layout

    def is_created(self) -> bool:
        return self._created

    @abstractmethod
    def create_layout(self, layout: PageLayout) -> None: ...

    # noinspection PyUnresolvedReferences
    @wraps(create_layout, updated=())
    def _do_create_layout(self, layout: PageLayout, *args, **kwargs) -> None:
        if self.is_registered():
            raise PageException(f"Trying to create layout of already registered {type(self).__name__}: {self.name}")
        if self.is_created():
            self._logger.warning(f"{type(self).__name__} '{self.id}' layout already created")
        else:
            self._at_create_layout(layout)
            self._run_create_layout(layout, *args, **kwargs)
            self._on_create_layout(layout)
            self._created = True

    def _at_create_layout(self, layout: PageLayout) -> None:
        pass

    def _on_create_layout(self, layout: PageLayout) -> None:
        pass

    def is_registered(self) -> bool:
        return self._registered

    # noinspection PyMethodMayBeStatic
    def register(self, **kwargs) -> None:
        self._logger.debug(f"Registering '{type(self).__name__}' page: {self.id} at {self.path}")
        dash.register_page(
            self.id,
            path=self.path,
            name=self.name,
            title=self.title,
            description=self.description,
            layout=self.layout.container,
            **kwargs,
        )

    # noinspection PyUnresolvedReferences
    @wraps(register, updated=())
    def _do_register(self, **kwargs) -> None:
        if not self.is_created():
            raise PageException(f"Trying to register {type(self).__name__} without layout: {self.name}")
        if self.is_registered():
            self._logger.warning(f"{type(self).__name__} '{self.id}' already registered")
            return

        self._logger.debug(f"Registering {type(self).__name__} '{self.id}': {self.name}")

        self._at_register(**kwargs)
        self._run_register(**kwargs)
        self._on_register(**kwargs)
        self._registered = True

    def _at_register(self, **kwargs) -> None:
        pass

    def _on_register(self, **kwargs) -> None:
        pass

    @staticmethod
    def _encode_id(id) -> str:
        # TODO: Implement correct URI encoding
        id = re.sub(r"[^0-9A-Za-zäöüÄÖÜß]+", "-", id).lower()
        id = id.replace("ä", "ae").replace("ö", "oe").replace("ü", "ue")
        return id


class PageException(InterfaceException):
    """
    Raise if an error occurred accessing a page.

    """



# -*- coding: utf-8 -*-
"""
lories.components.weather.dwd
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from . import brightsky  # noqa: F401
from .brightsky import Brightsky  # noqa: F401

from . import provider  # noqa: F401
from .provider import DeutscherWetterDienst  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories.data.database
~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import datetime as dt
from abc import abstractmethod
from functools import wraps
from typing import Any, Optional, overload

import tzlocal

import pandas as pd
import pytz as tz
from lories.connectors.connector import Connector, ConnectorMeta
from lories.connectors.errors import ConnectionError, ConnectorError
from lories.core import Configurations, Resources
from lories.core.typing import Timestamp
from lories.data.util import hash_data
from lories.data.validation import validate_index, validate_timezone
from lories.util import convert_timezone, to_date, to_timezone

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


class DatabaseMeta(ConnectorMeta):
    # noinspection PyProtectedMember
    def __call__(cls, *args, **kwargs):
        database = super().__call__(*args, **kwargs)
        cls._wrap_method(database, "hash")
        cls._wrap_method(database, "exists")
        cls._wrap_method(database, "read_first")
        cls._wrap_method(database, "read_first_index")
        cls._wrap_method(database, "read_last")
        cls._wrap_method(database, "read_last_index")
        cls._wrap_method(database, "delete")

        return database


# noinspection PyUnresolvedReferences
class Database(Connector, metaclass=DatabaseMeta):
    timezone: tz.BaseTzInfo

    def configure(self, configs: Configurations) -> None:
        super().configure(configs)

        timezone = configs.get("timezone", None)
        if timezone is None:
            timezone = tzlocal.get_localzone_name()
        self.timezone = to_timezone(timezone)

    # noinspection PyShadowingBuiltins
    def hash(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        method: Literal["MD5", "SHA1", "SHA256", "SHA512"] = "MD5",
        encoding: str = "UTF-8",
    ) -> Optional[str]:
        data = self._run_read(resources, start, end)
        data = self._validate(resources, data)
        data = self._get_range(data, start, end)
        if data is None or data.empty:
            return None

        columns = [r.id for r in resources if r.id in data.columns]
        return hash_data(data.loc[:, columns], method, encoding)

    @wraps(hash, updated=())
    def _do_hash(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        method: Literal["MD5", "SHA1", "SHA256", "SHA512"] = "MD5",
        encoding: str = "UTF-8",
        *args,
        **kwargs,
    ) -> Optional[str]:
        with self._lock:
            if not self._is_connected():
                raise ConnectionError(self, f"Database '{self.id}' not connected")

            return self._run_hash(resources, start=start, end=end, method=method, encoding=encoding, *args, **kwargs)

    def exists(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> bool:
        # TODO: Replace this placeholder more resource efficient
        data = self._run_read(resources, start, end)
        data = self._validate(resources, data)
        data = self._get_range(data, start, end)
        return not data.empty

    @wraps(exists, updated=())
    def _do_exists(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        *args,
        **kwargs,
    ) -> bool:
        with self._lock:
            if not self._is_connected():
                raise ConnectionError(self, f"Database '{self.id}' not connected")

            return self._run_exists(resources, start=start, end=end, *args, **kwargs)

    @overload
    def read(self, resources: Resources) -> pd.DataFrame: ...

    @overload
    def read(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> pd.DataFrame: ...

    @abstractmethod
    def read(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> pd.DataFrame: ...

    @wraps(read, updated=())
    def _do_read(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        with self._lock:
            if not self._is_connected():
                raise ConnectorError(self, f"Trying to read from unconnected {type(self).__name__}: {self.id}")

            data = self._run_read(resources, start=start, end=end, *args, **kwargs)
            data = self._validate(resources, data)
            return self._get_range(data, start, end)

    @abstractmethod
    def read_first(self, resources: Resources) -> Optional[pd.DataFrame]: ...

    @wraps(read_first, updated=())
    def _do_read_first(self, resources: Resources, *args, **kwargs) -> Optional[pd.DataFrame]:
        with self._lock:
            if not self._is_connected():
                raise ConnectionError(self, f"Database '{self.id}' not connected")

            data = self._run_read_first(resources, *args, **kwargs)
            data = self._validate(resources, data)
            return data

    def read_first_index(self, resources: Resources) -> Optional[Any]:
        data = self._run_read_first(resources)
        data = self._validate(resources, data)
        if data is None or data.empty:
            return None
        return min(data.index)

    @wraps(read_first_index, updated=())
    def _do_read_first_index(self, resources: Resources, *args, **kwargs) -> Optional[Any]:
        with self._lock:
            if not self._is_connected():
                raise ConnectionError(self, f"Database '{self.id}' not connected")

            index = self._run_read_first_index(resources, *args, **kwargs)
            if isinstance(index, (pd.Timestamp, dt.datetime)):
                index = convert_timezone(index, timezone=self.timezone)
            return index

    @abstractmethod
    def read_last(self, resources: Resources) -> Optional[pd.DataFrame]: ...

    @wraps(read_last, updated=())
    def _do_read_last(self, resources: Resources, *args, **kwargs) -> Optional[pd.DataFrame]:
        with self._lock:
            if not self._is_connected():
                raise ConnectionError(self, f"Database '{self.id}' not connected")

            data = self._run_read_last(resources, *args, **kwargs)
            data = self._validate(resources, data)
            return data

    def read_last_index(self, resources: Resources) -> Optional[pd.Index]:
        data = self._run_read_last(resources)
        data = self._validate(resources, data)
        if data is None or data.empty:
            return None
        return max(data.index)

    @wraps(read_last_index, updated=())
    def _do_read_last_index(self, resources: Resources, *args, **kwargs) -> Optional[Any]:
        with self._lock:
            if not self._is_connected():
                raise ConnectionError(self, f"Database '{self.id}' not connected")

            index = self._run_read_last_index(resources, *args, **kwargs)
            if isinstance(index, (pd.Timestamp, dt.datetime)):
                index = convert_timezone(index, timezone=self.timezone)
            return index

    def _validate(self, resources: Resources, data: pd.DataFrame) -> pd.DataFrame:
        if not data.empty:
            data = validate_index(data)
            data.index = validate_timezone(data.index, self.timezone)
            for resource in resources:
                if resource.id not in data:
                    continue
                if resource.type in [pd.Timestamp, dt.datetime]:
                    resource_data = data[resource.id]
                    if pd.api.types.is_string_dtype(resource_data.values):
                        resource_data = pd.to_datetime(resource_data)
                    data[resource.id] = validate_timezone(resource_data, self.timezone)
        return data

    @staticmethod
    def _get_range(
        data: pd.DataFrame,
        start: Optional[Timestamp | str] = None,
        end: Optional[Timestamp | str] = None,
        **kwargs,
    ) -> pd.DataFrame:
        if data.empty:
            return data
        if start is not None:
            start = to_date(start, **kwargs)
            data = data[data.index >= start]
        if end is not None:
            end = to_date(end, **kwargs)
            data = data[data.index <= end]
        return data

    @overload
    def delete(self, resources: Resources) -> None: ...

    @overload
    def delete(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> None: ...

    def delete(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> None:
        raise NotImplementedError(f"Unable to delete values for database '{self.id}'")

    @wraps(delete, updated=())
    def _do_delete(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        *args,
        **kwargs,
    ) -> None:
        with self._lock:
            if not self._is_connected():
                raise ConnectionError(self, f"Database '{self.id}' not connected")

            self._run_delete(resources, start=start, end=end, *args, **kwargs)



# -*- coding: utf-8 -*-
"""
lories.core.activator
~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from functools import wraps
from typing import Callable, Dict

from lories._core._activator import Activator as ActivatorType  # noqa
from lories._core._activator import _Activator  # noqa
from lories.core.configs import ConfigurationError
from lories.core.configs.configurator import Configurator, ConfiguratorMeta


class ActivatorMeta(ConfiguratorMeta):
    # noinspection PyProtectedMember
    def __call__(cls, *args, **kwargs):
        activator = super().__call__(*args, **kwargs)
        cls._wrap_method(activator, "activate")
        cls._wrap_method(activator, "deactivate")

        return activator


# noinspection PyAbstractClass
class Activator(_Activator, Configurator, metaclass=ActivatorMeta):
    _active: bool = False

    def __enter__(self) -> ActivatorType:
        self.activate()
        return self

    # noinspection PyShadowingBuiltins
    def __exit__(self, type, value, traceback) -> None:
        self.deactivate()

    # noinspection PyShadowingBuiltins
    def _convert_vars(self, convert: callable = str) -> Dict[str, str]:
        values = super()._convert_vars(convert)
        values["active"] = str(self.is_active())
        return values

    def is_active(self) -> bool:
        return self._active

    def activate(self) -> None:
        pass

    # noinspection PyUnresolvedReferences
    @wraps(activate, updated=())
    def _do_activate(self, *args, **kwargs) -> None:
        if not self.is_enabled():
            raise ConfigurationError(f"Trying to activate disabled '{type(self).__name__}': {self.id}")
        if not self.is_configured():
            raise ConfigurationError(f"Trying to activate unconfigured '{type(self).__name__}': {self.id}")
        if self.is_active():
            self._logger.warning(f"Trying to activate already active '{type(self).__name__}': {self.id}")
            return

        self._at_activate()
        self._run_activate(*args, **kwargs)
        self._on_activate()
        self._active = True

    def _at_activate(self) -> None:
        pass

    def _on_activate(self) -> None:
        pass

    def deactivate(self) -> None:
        pass

    # noinspection PyUnresolvedReferences
    @wraps(deactivate, updated=())
    def _do_deactivate(self, *args, **kwargs) -> None:
        if not self.is_active():
            return

        self._at_deactivate()
        self._run_deactivate(*args, **kwargs)
        self._on_deactivate()
        self._active = False

    def _at_deactivate(self) -> None:
        pass

    def _on_deactivate(self) -> None:
        pass


def activating(method: Callable[..., None]) -> Callable[..., None]:
    @wraps(method)
    def _activating(activator: Activator, *args, **kwargs):
        if activator is None or not isinstance(activator, Activator):
            raise TypeError(f"Method '{method}' not of Activator class")
        was_activated = False
        if not activator.is_active():
            activator.activate()
            was_activated = True
        try:
            method(activator, *args, **kwargs)

        finally:
            if was_activated:
                activator.deactivate()

    return _activating



# -*- coding: utf-8 -*-
"""
lories.connectors.connector
~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import datetime as dt
from collections import OrderedDict
from functools import wraps
from threading import Lock
from typing import Any, Dict, Optional

import pandas as pd
import pytz as tz
from lories._core._configurations import Configurations  # noqa
from lories._core._connector import ConnectType, _Connector  # noqa
from lories._core._context import _Context  # noqa
from lories._core._registrator import RegistratorContext  # noqa
from lories.connectors.errors import ConnectorError
from lories.core import Resource, ResourceError, Resources
from lories.core.configs.configurator import Configurator, ConfiguratorMeta
from lories.core.configs.errors import ConfigurationError
from lories.core.register.registrator import Registrator
from lories.data.channels import Channel, Channels, ChannelState
from lories.data.validation import validate_index


class ConnectorMeta(ConfiguratorMeta):
    # noinspection PyProtectedMember
    def __call__(cls, *args, **kwargs):
        connector = super().__call__(*args, **kwargs)
        cls._wrap_method(connector, "connect")
        cls._wrap_method(connector, "disconnect")
        cls._wrap_method(connector, "read")
        cls._wrap_method(connector, "write")

        return connector


# noinspection PyAbstractClass
class Connector(_Connector, Registrator, metaclass=ConnectorMeta):
    _connected: bool = False
    _connect_type: ConnectType = ConnectType.AUTO

    _timestamp_connect: pd.Timestamp = pd.NaT
    _timestamp_disconnect: pd.Timestamp = pd.NaT
    _interval_reconnect: pd.Timedelta = pd.Timedelta(minutes=1)

    __resources: Resources

    _lock: Lock

    def __init__(
        self,
        context: RegistratorContext,
        configs: Optional[Configurations] = None,
        **kwargs,
    ) -> None:
        super().__init__(context=context, configs=configs, **kwargs)
        self.__resources = Resources()
        self._lock = Lock()

    def __enter__(self) -> Connector:
        self.connect(self.__resources)
        return self

    # noinspection PyShadowingBuiltins
    def __exit__(self, type, value, traceback):
        self.disconnect()

    @classmethod
    def _assert_context(cls, context: RegistratorContext) -> RegistratorContext:
        if context is None:
            raise TypeError(f"Invalid '{cls.__name__}' context: {type(context)}")
        return super()._assert_context(context)

    # noinspection PyShadowingBuiltins, PyProtectedMember
    def _get_vars(self) -> Dict[str, Any]:
        vars = super()._get_vars()

        # Channels are a subset of resources, hence omit them from printing
        vars.pop("channels", None)
        return vars

    # noinspection PyShadowingBuiltins
    def _convert_vars(self, convert: callable = str) -> Dict[str, str]:
        vars = self._get_vars()
        values = OrderedDict()
        try:
            id = vars.pop("id", self.id)
            key = vars.pop("key", self.key)
            if id != key:
                values["id"] = id
            values["key"] = key
        except (ResourceError, AttributeError):
            # Abstract properties are not yet instanced
            pass

        if "name" in vars:
            values["name"] = vars.pop("name")

        values.update(
            {
                k: str(v) if not isinstance(v, (_Context, Configurator, Resource, Resources)) else convert(v)
                for k, v in vars.items()
            }
        )
        values["context"] = convert(self.context)
        values["configurations"] = convert(self.configs)
        values["configured"] = str(self.is_configured())
        values["connected"] = str(self._is_connected())
        values["enabled"] = str(self.is_enabled())
        return values

    @property
    def resources(self) -> Resources:
        return self.__resources

    @property
    def channels(self) -> Channels:
        return Channels([resource for resource in self.__resources if isinstance(resource, Channel)])

    def set_channels(self, state: ChannelState) -> None:
        # Set only channel states for channels, that actively are getting read or written by this connector.
        # Local channels may be logging channels as well, which need to be skipped.
        for channel in self.channels.filter(lambda c: c.has_connector(self.id)):
            channel.state = state

    def configure(self, configs: Configurations) -> None:
        super().configure(configs)
        self._connect_type = ConnectType.get(configs.get("connect", default=True))

    def _is_disconnected(self) -> bool:
        return not self._is_connected()

    # noinspection SpellCheckingInspection
    def _is_reconnectable(self) -> bool:
        if not self.is_enabled() or not self.is_configured() or not self._is_connectable():
            return False
        if not pd.isna(self._timestamp_connect):
            return self._timestamp_connect + self._interval_reconnect <= pd.Timestamp.now(tz.UTC)
        if not pd.isna(self._timestamp_disconnect):
            return self._timestamp_disconnect + self._interval_reconnect <= pd.Timestamp.now(tz.UTC)
        return False

    def _is_connectable(self) -> bool:
        return self._is_disconnected() and self._connect_type == ConnectType.AUTO

    def _is_connected(self) -> bool:
        return self.is_connected() and self._connected

    def is_connected(self) -> bool:
        return True

    # noinspection PyUnresolvedReferences, PyTypeChecker
    @wraps(_Connector.connect, updated=())
    def _do_connect(self, resources: Resources, *args, **kwargs) -> None:
        with self._lock:
            if not self.is_enabled():
                raise ConfigurationError(f"Trying to connect disabled {type(self).__name__}: {self.id}")
            if not self.is_configured():
                raise ConfigurationError(f"Trying to connect unconfigured {type(self).__name__}: {self.id}")

            self._timestamp_connect = pd.Timestamp.now(tz.UTC)
            self._timestamp_disconnect = pd.NaT

            if not self._is_connected():
                self._at_connect(resources)
                self._run_connect(resources, *args, **kwargs)
                self._on_connect(resources)
                self.__resources = resources
            else:
                self._logger.warning(f"{type(self).__name__} '{self.id}' already connected")

            self._connected = True

    def _at_connect(self, resources: Resources) -> None:
        pass

    def _on_connect(self, resources: Resources) -> None:
        pass

    # noinspection PyUnresolvedReferences, PyTypeChecker
    @wraps(_Connector.disconnect, updated=())
    def _do_disconnect(self) -> None:
        with self._lock:
            self._timestamp_connect = pd.NaT
            self._timestamp_disconnect = pd.Timestamp.now(tz.UTC)

            if not self._is_disconnected():
                self._at_disconnect()
                self._run_disconnect()
                self._on_disconnect()

            self._connected = False

    def _at_disconnect(self) -> None:
        pass

    def _on_disconnect(self) -> None:
        pass

    # noinspection PyUnresolvedReferences, PyTypeChecker
    @wraps(_Connector.read, updated=())
    def _do_read(self, resources: Resources, *args, **kwargs) -> pd.DataFrame:
        with self._lock:
            if not self._is_connected():
                raise ConnectorError(self, f"Trying to read from unconnected {type(self).__name__}: {self.id}")

            data = self._run_read(resources, *args, **kwargs)
            data = self._validate(resources, data)
            return data

    # noinspection PyMethodMayBeStatic
    def _validate(self, resources: Resources, data: pd.DataFrame) -> pd.DataFrame:
        if not data.empty:
            data = validate_index(data)
            for resource in resources:
                if resource.id not in data:
                    continue
                if resource.type in [pd.Timestamp, dt.datetime]:
                    resource_data = data[resource.id]
                    if pd.api.types.is_string_dtype(resource_data.values):
                        data[resource.id] = pd.to_datetime(resource_data)
        return data

    # noinspection PyUnresolvedReferences, PyTypeChecker
    @wraps(_Connector.write, updated=())
    def _do_write(self, data: pd.DataFrame, *args, **kwargs) -> None:
        with self._lock:
            if not self._is_connected():
                raise ConnectorError(self, f"Trying to write to unconnected {type(self).__name__}: {self.id}")
            unknown = [c for c in data.columns if c not in self.resources]
            if len(unknown) > 0:
                raise ConnectorError(
                    self,
                    f"Trying to write unknown resource{'s' if len(unknown) > 0 else ''} '{', '.join(unknown)}' for "
                    f"{type(self).__name__}: {self.id}",
                )

            self._run_write(data, *args, **kwargs)



# -*- coding: utf-8 -*-
"""
lories.connectors.virtual
~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import random
from typing import Optional

import pandas as pd
import pytz as tz
from lories.connectors import Connector, ConnectorError, register_connector_type
from lories.core.configs import ConfigurationError
from lories.typing import Channel, Resource, Resources, Timestamp


# noinspection PyShadowingBuiltins
@register_connector_type("virtual", "random", "dummy")
class VirtualConnector(Connector):
    VIRTUAL: str = "virtual"
    RANDOM: str = "random"

    _data: pd.Series

    def connect(self, resources: Resources) -> None:
        super().connect(resources)
        index = []
        data = []
        for resource in resources:
            index.append(resource.id)

            generator = resource.get("generator", default=None)
            if generator == VirtualConnector.RANDOM:
                for attr in ["min", "max"]:
                    if attr not in resource:
                        raise ConfigurationError(f"Invalid dummy channel '{resource.id}', missing attribute: {attr}")
                data.append(float(random.randrange(int(resource.min * 100), int(resource.max * 100))) / 100.0)

            elif generator is not None:
                raise ConfigurationError(f"Invalid dummy channel '{resource.id}' generator: {generator}")
            else:
                data.append(resource.get("default", default=None))
        self._data = pd.Series(index=index, data=data)

    def read(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> pd.DataFrame:
        for resource in resources:
            generator = resource.get("generator", default=VirtualConnector.VIRTUAL)
            if generator == VirtualConnector.RANDOM:
                self._read_random(resource)

            elif generator != VirtualConnector.VIRTUAL:
                raise ConnectorError(self, f"Trying to read dummy channel '{resource.id}' with generator: {generator}")
        return self._data.to_frame(pd.Timestamp.now(tz.UTC).floor(freq="s")).T

    def _read_random(self, resource: Resource) -> None:
        range = int(abs(resource.max - resource.min))
        value = float(random.randrange(-range * 100, range * 100)) / 1000.0 + self._data[resource.id]
        if value < resource.min:
            value = resource.min
        if value > resource.max:
            value = resource.max
        self._data[resource.id] = value

    def write(self, data: pd.DataFrame) -> None:
        for id in data.columns:
            if id in self.channels:
                channel = self.channels[id]
                generator = channel.get("generator", default=VirtualConnector.VIRTUAL)
                if generator == VirtualConnector.VIRTUAL:
                    self._write_virtual(data, channel)

                elif generator == VirtualConnector.RANDOM:
                    self._write_random(data, channel)

                else:
                    raise ConnectorError(
                        self, f"Trying to write to dummy channel '{channel.id}' with generator: {generator}"
                    )

    def _write_virtual(self, data: pd.DataFrame, channel: Channel) -> None:
        self._data[channel.id] = data.at[data.index[-1], channel.id]

    def _write_random(self, data: pd.DataFrame, channel: Channel) -> None:
        value = data.at[data.index[-1], channel.id]
        if value < channel.min:
            value = channel.min
        if value > channel.max:
            value = channel.max
        self._data[channel.id] = value



# -*- coding: utf-8 -*-
"""
lories.data.converters.io.analog
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Any, Optional

from lories.core import ConfigurationError
from lories.core.typing import Configurations
from lories.data.converters import ConversionError, register_converter_type
from lories.data.converters.converter import FloatConverter


@register_converter_type("analog_input")
class AnalogInput(FloatConverter):
    INPUT_KEY: str = "input"

    _input_max: float
    _input_min: float
    _input_zero: float

    _invert: bool
    _factor: float
    _divisor: float

    max: float
    min: float

    def configure(self, configs: Configurations) -> None:
        super().configure(configs)
        self.min = configs.get_float("value_min", default=0.0)
        self.max = configs.get_float("value_max")
        if self.max <= self.min:
            raise ConfigurationError(
                f"Invalid value boundaries for '{self.id}' with min ({self.min}) > max ({self.max})"
            )
        self._configure_input(configs)
        self._assert_input(self._input_min, self._input_max, self._input_zero)

        self._invert = configs.get_bool("invert", default=False)
        self._factor = configs.get_float("factor", default=1)
        self._divisor = (self._input_max - self._input_min) / (self.max - self.min)

    def _configure_input(
        self,
        configs: Configurations,
        default_max: float = None,
        default_min: float = 0.0,
    ) -> None:
        self._input_max = configs.get_float(f"{self.INPUT_KEY}_max", default=default_max)
        self._input_min = configs.get_float(f"{self.INPUT_KEY}_min", default=default_min)
        self._input_zero = configs.get_float(f"{self.INPUT_KEY}_zero", default=self._input_min)

    # noinspection PyShadowingBuiltins
    def _assert_input(self, min: float, max: float, zero: float) -> None:
        if min is None:
            raise ConfigurationError(f"Missing input minimum for '{self.id}'")
        if max is None:
            raise ConfigurationError(f"Missing input maximum for '{self.id}'")
        if max <= min:
            raise ConfigurationError(f"Invalid input boundaries for '{self.id}' with min ({min}) > max ({max})")

        if zero is None:
            raise ConfigurationError(f"Missing input zero point for '{self.id}'")
        if min > zero > max:
            raise ConfigurationError(f"Invalid input zero point for '{self.id}': {zero}")

    def convert(self, value: Any, **kwargs) -> Optional[float]:
        if self._input_min > value > self._input_max:
            raise ConversionError(
                f"Invalid input signal out of limit ({self._input_min} to {self._input_max}): " + str(value)
            )
        _value = (value * self._factor - self._input_zero) / self._divisor

        if self._invert:
            _value = self.max - _value
        return _value



# -*- coding: utf-8 -*-
"""
lories.application
~~~~~~~~~~~~~~~~~~


"""

from .interface import (  # noqa: F401
    Interface,
    InterfaceException,
    register_interface_type,
)

from .main import Application  # noqa: F401

import importlib

for import_interface in ["view"]:
    try:
        importlib.import_module(f".{import_interface}", "lories.application")

    except ModuleNotFoundError:
        # TODO: Implement meaningful logging here
        pass

del importlib



# -*- coding: utf-8 -*-
"""
lories.connectors.tasks.check
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Optional

from lories._core._database import _Database  # noqa
from lories._core.typing import Timestamp  # noqa
from lories.connectors.tasks.task import ConnectorTask


class CheckTask(ConnectorTask):
    def run(
        self,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> bool:
        self._logger.debug(
            f"Checking data for {len(self.channels)} channels of '{type(self.connector).__name__}': {self.connector.id}"
        )
        if isinstance(self.connector, _Database):
            return self.connector.exists(self.channels, start=start, end=end)
        else:
            return False



# -*- coding: utf-8 -*-
"""
lories.data.listeners
~~~~~~~~~~~~~~~~~~~~~


"""

from .listener import (  # noqa: F401
    Listener,
    ListenerError,
)

from .context import ListenerContext  # noqa: F401



"""
lories.core.typing
~~~~~~~~~~~~~~~~~~


"""

from typing import TypeAlias

from ..._core.typing import (  # noqa
    Timestamp,
    Timezone,
)

from ..._core._entity import Entity  # noqa
from ..._core._context import Context  # noqa

from ..._core._resource import Resource  # noqa
from ..._core._resources import (  # noqa
    Resources,
    ResourcesArgument,
)

from ..._core._channel import Channel  # noqa
from ..._core._channels import (  # noqa
    Channels,
    ChannelsArgument,
)

from ..._core._converter import (  # noqa
    Converter,
    Converters,
)

from ..._core._listener import Listener  # noqa

from ..._core._configurations import Configurations  # noqa
from ..._core._configurator import Configurator  # noqa

from ..._core._registrator import (  # noqa
    Registrator,
    RegistratorContext,
)

from ..._core._activator import Activator  # noqa

from ..._core._connector import (  # noqa
    Connector,
    Connectors,
)
from ..._core._database import Database  # noqa

from ..._core._data import Data  # noqa

from ..._core._component import (  # noqa
    Component,
    Components,
)

ContextArgument: TypeAlias = RegistratorContext


__all__ = [
    "Timestamp",
    "Timezone",
    "Entity",
    "Context",
    "ContextArgument",
    "Resource",
    "Resources",
    "ResourcesArgument",
    "Channel",
    "Channels",
    "ChannelsArgument",
    "Converter",
    "Converters",
    "Listener",
    "Configurations",
    "Configurator",
    "Registrator",
    "Activator",
    "Connector",
    "Connectors",
    "Database",
    "Data",
    "Component",
    "Components",
]



# -*- coding: utf-8 -*-
"""
lories.data.converters.io
~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from . import analog  # noqa: F401
from .analog import AnalogInput  # noqa: F401

from . import current  # noqa: F401
from .current import CurrentInput  # noqa: F401

from . import voltage  # noqa: F401
from .voltage import VoltageInput  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories.data.converters.context
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Callable, Collection, List, Optional, Type

from lories._core._converter import _ConverterContext  # noqa
from lories.core import Configurations, ResourceError
from lories.core.register import RegistratorContext, Registry
from lories.data.converters.converter import (
    BoolConverter,
    BytesConverter,
    Converter,
    DatetimeConverter,
    FloatConverter,
    IntConverter,
    StringConverter,
    TimestampConverter,
)

CONVERTERS_BUILTINS = [
    DatetimeConverter,
    TimestampConverter,
    StringConverter,
    FloatConverter,
    IntConverter,
    BoolConverter,
    BytesConverter,
]

registry = Registry[Converter]()
registry.register(DatetimeConverter, "datetime")
registry.register(TimestampConverter, "timestamp")
registry.register(StringConverter, "str", "string")
registry.register(FloatConverter, "float")
registry.register(IntConverter, "int", "integer")
registry.register(BoolConverter, "bool", "boolean")
registry.register(BytesConverter, "byte", "bytes")


def register_converter_type(
    key: str,
    *alias: str,
    factory: Callable[[ConverterContext, Optional[Configurations]], Converter] = None,
    replace: bool = False,
) -> Callable[[Type[Converter]], Type[Converter]]:
    # noinspection PyShadowingNames
    def _register(cls: Type[Converter]) -> Type[Converter]:
        registry.register(cls, key, *alias, factory=factory, replace=replace)
        return cls

    return _register


class ConverterContext(_ConverterContext, RegistratorContext[Converter]):
    @property
    def _registry(self) -> Registry[Converter]:
        return registry

    # noinspection PyProtectedMember
    def load(
        self,
        configs: Optional[Configurations] = None,
        configure: bool = True,
        **kwargs,
    ) -> Collection[Converter]:
        if configs is None:
            configs = self._load_registrators_configs()
        defaults = Converter._build_defaults(configs)

        converters = []
        converter_dirs = configs.dirs.to_dict()
        converter_dirs["conf_dir"] = str(configs.dirs.conf.joinpath("converters.d"))
        converter_generics = [c.type for c in registry.filter(lambda r: r.type in CONVERTERS_BUILTINS)]
        for converter_cls in converter_generics:
            converter_key = converter_cls.dtype.__name__.lower()
            converter_configs = Configurations.load(
                f"{converter_key}.conf",
                require=False,
                **converter_dirs,
                **defaults,
            )
            converter = converter_cls(context=self, configs=converter_configs, key=converter_key)
            converters.append(converter)
            self._add(converter)

        converters.extend(self._load(self, configs, includes=Converter.INCLUDES, configure=False, **kwargs))
        if configure:
            self.configure(converters)
        return converters

    def has_dtype(self, *dtypes: Type) -> bool:
        return len(self._get_by_dtypes(*dtypes)) > 0

    def get_by_dtype(self, dtype: Type) -> Converter:
        converters = self._get_by_dtypes(dtype)
        if len(converters) == 0:
            raise ResourceError(f"Converter instance for '{dtype}' does not exist")
        return converters[0]

    def _get_by_dtypes(self, dtype: Type) -> List[Converter]:
        return [t for t in self.values() if issubclass(t.dtype, dtype)]



# -*- coding: utf-8 -*-
"""
lories.util
~~~~~~~~~~~


"""

from __future__ import annotations

import datetime as dt
import dateutil.parser
import re
from dateutil.relativedelta import relativedelta
from pydoc import locate
from typing import Any, Callable, Collection, Dict, List, Mapping, Optional, Tuple, Type, TypeVar

import tzlocal

import numpy as np
import pandas as pd
import pytz as tz
from lories._core._configurations import Configurations  # noqa
from lories._core.typing import Timestamp, Timezone  # noqa
from lories.core.errors import ResourceError

# noinspection SpellCheckingInspection
INVALID_CHARS = "'!@#$%^&?*;:,./\\|`´+~=- "

C = TypeVar("C")  # , bound=Context)
V = TypeVar("V")


# noinspection PyShadowingBuiltins
def get_context(object: Any, type: Type[C] | Collection[Type[C]]) -> Optional[C]:
    _context = object
    if _context is None or isinstance(_context, type):
        return _context
    while not isinstance(_context, type):
        try:
            _context = _context.context
        except AttributeError:
            raise TypeError(f"Invalid context type: {object.__class__}")
    return _context


# noinspection PyShadowingBuiltins
def get_includes(type: Type) -> List[str]:
    includes = []
    for _type in type.mro():
        if not hasattr(_type, "INCLUDES"):
            continue
        _includes = getattr(_type, "INCLUDES")
        if not isinstance(_includes, Collection) or isinstance(_includes, str):
            continue
        for include in _includes:
            if include not in includes:
                includes.append(include)
    return includes


# noinspection PyShadowingBuiltins, PyShadowingNames
def get_variables(
    object: Any,
    include: Optional[Type[V]] = object,
    exclude: Optional[Type | Tuple[Type, ...]] = None,
) -> List[V]:
    def _is_type(o) -> bool:
        return isinstance(o, include) and (exclude is None or not isinstance(o, exclude))

    if isinstance(object, Collection):
        return [o for o in object if _is_type(o)]
    return list(get_members(object, lambda attr, member: _is_type(member)).values())


# noinspection PyShadowingBuiltins
def get_members(
    object: Any,
    filter: Optional[Callable] = None,
    private: bool = False,
) -> Dict[str, Any]:
    members = dict()
    processed = set()
    for attr in dir(object):
        try:
            member = getattr(object, attr)
            # Handle duplicate attr
            if attr in processed:
                raise AttributeError
        except (AttributeError, ResourceError):
            continue
        if (
            (private or "__" not in attr)
            and (filter is None or filter(attr, member))
            and member not in members.values()
        ):
            members[attr] = member
        processed.add(attr)
    return dict(sorted(members.items()))


# noinspection PyTypeChecker
def update_recursive(
    configs: Dict[str, Any] | Configurations,
    update: Mapping[str, Any] | Configurations,
    replace: bool = True,
) -> Dict[str, Any]:
    for key, update_value in update.items():
        if isinstance(update_value, Mapping):
            update_map = configs.get(key, {})
            if isinstance(update_map, Mapping):
                update_map = update_recursive(update_map, update_value, replace=replace)
            elif not replace:
                continue
            configs[key] = update_map
        elif key not in configs.keys() or replace:
            configs[key] = update_value
    return configs


def convert_timezone(
    date: Optional[Timestamp | str],
    timezone: Optional[Timezone] = None,
) -> Optional[pd.Timestamp]:
    if date is None:
        return None
    if isinstance(date, str):
        date = dateutil.parser.parse(date)
    if isinstance(date, dt.datetime):
        date = pd.Timestamp(date)

    if isinstance(date, pd.Timestamp):
        if timezone is None:
            timezone = to_timezone(tzlocal.get_localzone_name())
        if date.tzinfo is None or date.tzinfo.utcoffset(date) is None:
            return date.tz_localize(timezone)
        else:
            return date.tz_convert(timezone)
    else:
        raise TypeError(f"Unable to convert date of type {type(date)}")


def slice_range(
    start: Optional[Timestamp | str],
    end: Optional[Timestamp | str],
    timezone: Optional[Timezone] = None,
    freq: str = "D",
    **kwargs,
) -> List[Tuple[Optional[pd.Timestamp], Optional[pd.Timestamp]]]:
    start = to_date(start, timezone, **kwargs)
    end = to_date(end, timezone, **kwargs)

    if start is None or end is None:
        return [(start, end)]

    freq_delta = to_timedelta(freq)

    def _next(timestamp: pd.Timestamp) -> pd.Timestamp:
        _timestamp = floor_date(timestamp + freq_delta, freq=freq)
        if _timestamp > timestamp:
            return _timestamp
        # Handle daylight savings
        return floor_date(timestamp + freq_delta * 2, freq=freq)

    ranges = []
    range_start = start
    if floor_date(start, freq=freq) == start:
        range_start += pd.Timedelta(seconds=1)
    range_end = min(_next(start), end)
    ranges.append((range_start, range_end))

    while range_end < end:
        range_start = _next(range_start)
        range_end = min(_next(range_start), end)
        ranges.append((range_start + pd.Timedelta(seconds=1), range_end))
    return ranges


def floor_date(
    date: Optional[Timestamp | str],
    timezone: Optional[Timezone] = None,
    freq: str = "D",
) -> Optional[pd.Timestamp]:
    if date is None:
        return None
    date = to_date(date)
    if timezone is None:
        timezone = date.tzinfo
    date = convert_timezone(date, timezone)
    freq = parse_freq(freq)
    if any([freq.endswith(f) for f in ["Y", "M", "W"]]):
        return date.tz_localize(None).to_period(freq).to_timestamp().tz_localize(timezone, ambiguous=True)
    elif any([freq.endswith(f) for f in ["D", "h", "min", "s"]]):
        return date.tz_localize(None).floor(freq).tz_localize(timezone, ambiguous=True)
    else:
        raise ValueError(f"Invalid frequency: {freq}")


# noinspection PyTypeChecker
def ceil_date(
    date: Optional[Timestamp | str],
    timezone: Optional[Timezone] = None,
    freq: str = "D",
) -> Optional[pd.Timestamp]:
    if date is None:
        return None
    date = floor_date(date, timezone, freq)

    return date + to_timedelta(freq) - pd.Timedelta(microseconds=1)


# noinspection PyShadowingBuiltins
def to_date(
    date: Optional[Timestamp | str | int],
    timezone: Optional[Timezone] = None,
    format: Optional[str] = None,
) -> Optional[pd.Timestamp]:
    if date is None:
        return None

    if issubclass(type(date), dt.datetime):
        return convert_timezone(date, timezone)
    if isinstance(date, int):
        return convert_timezone(pd.Timestamp(date, unit="s"), timezone)
    if isinstance(date, str):
        if format is None:
            return convert_timezone(pd.Timestamp(date), timezone)
        return convert_timezone(pd.Timestamp(dt.datetime.strptime(date, format)), timezone)

    raise TypeError(f"Invalid date type: {type(date)}")


def to_timezone(timezone: Optional[Timezone | str | int | float]) -> Optional[Timezone]:
    if timezone is None:
        return None
    if isinstance(timezone, (tz.BaseTzInfo, dt.tzinfo)):
        return timezone
    if isinstance(timezone, str):
        try:
            return tz.timezone(timezone)

        except tz.UnknownTimeZoneError:
            # Handle the case where the timezone is not recognized
            if timezone == "CEST":
                return tz.FixedOffset(2 * 60)

            # Handle offset time strings
            return pd.to_datetime(timezone, format="%z").tzinfo

    if isinstance(timezone, (int, float)):
        return tz.FixedOffset(timezone * 60)

    raise TypeError(f"Invalid timezone type: {type(timezone)}")


def to_timedelta(freq: str) -> relativedelta | pd.Timedelta:
    freq_val = "".join(s for s in freq if s.isnumeric())
    freq_val = int(freq_val) if len(freq_val) > 0 else 1
    freq = parse_freq(freq)
    if freq.endswith("Y"):
        return relativedelta(years=freq_val)
    elif freq.endswith("M"):
        return relativedelta(months=freq_val)
    elif freq.endswith("W"):
        return relativedelta(weeks=freq_val)
    elif freq.endswith("D"):
        return pd.Timedelta(days=freq_val)
    elif freq.endswith("h"):
        return pd.Timedelta(hours=freq_val)
    elif freq.endswith("min"):
        return pd.Timedelta(minutes=freq_val)
    elif freq.endswith("s"):
        return pd.Timedelta(seconds=freq_val)
    else:
        raise ValueError(f"Invalid frequency: {freq}")


def is_float(value: str | float) -> bool:
    def _is_numeric(s: str) -> bool:
        return s.replace(".", "").replace(",", "").replace("e+", "").replace("e-", "").replace("e", "").isnumeric()

    if (
        issubclass(type(value), (np.integer, int))
        or issubclass(type(value), (np.floating, float))
        or (isinstance(value, str) and _is_numeric(value))
    ):
        return True
    return False


def to_float(value: str | float) -> Optional[float]:
    if value is None:
        return None
    if type(value) == float:  # noqa E721
        return value
    if is_float(value):
        return float(value)
    raise TypeError(f"Expected str, int or float, not: {type(value)}")


def is_int(value: str | int) -> bool:
    if (
        (issubclass(type(value), (np.floating, float)) and int(value) == value)
        or issubclass(type(value), (np.integer, int))
        or (isinstance(value, str) and value.isnumeric())
    ):
        return True
    return False


def to_int(value: str | int) -> Optional[int]:
    if value is None:
        return None
    if type(value) == int:  # noqa E721
        return value
    if is_int(value):
        return int(value)
    raise TypeError(f"Expected str, float or int, not: {type(value)}")


def is_bool(value: str | bool) -> bool:
    if issubclass(type(value), (np.bool, bool)):
        return True
    if isinstance(value, str) and any(value.lower() == b for b in ["true", "false", "yes", "no", "y", "n"]):
        return True
    return False


def to_bool(value: str | bool) -> Optional[bool]:
    if value is None:
        return None
    if type(value) == bool:  # noqa E721
        return value
    if isinstance(value, str):
        if value.lower() in ["true", "yes", "y"]:
            return True
        if value.lower() in ["false", "no", "n"]:
            return False
    if issubclass(type(value), (np.bool, bool, np.integer, int)):
        return bool(value)
    raise TypeError(f"Expected str or bool, not '{type(value)}': {value}")


# noinspection SpellCheckingInspection
def parse_freq(freq: str) -> Optional[str]:
    if freq is None:
        return None
    match = re.fullmatch(r"\s*(\d*)\s*([a-zA-Z]+)\s*", freq)
    if not match:
        raise ValueError(f"Invalid frequency format: '{freq}'")

    number_part, unit_part = match.groups()
    value = int(number_part) if number_part else 1

    def _parse_freq(suffix: str) -> str:
        return str(value) + suffix if value > 1 else suffix

    if unit_part.lower() in ["y", "year", "years"]:
        return _parse_freq("Y")
    elif unit_part.lower() in ["m", "month", "months"]:
        return _parse_freq("M")
    elif unit_part.lower() in ["w", "week", "weeks"]:
        return _parse_freq("W")
    elif unit_part.lower() in ["d", "day", "days"]:
        return _parse_freq("D")
    elif unit_part.lower() in ["h", "hour", "hours"]:
        return _parse_freq("h")
    elif unit_part.lower() in ["t", "min", "mins"]:
        return _parse_freq("min")
    elif unit_part.lower() in ["s", "sec", "secs"]:
        return _parse_freq("s")
    else:
        raise ValueError(f"Invalid frequency: {freq}")


# noinspection PyShadowingBuiltins
def parse_type(t: str | Type, default: Type = None) -> Type:
    if t is None:
        if default is None:
            raise ValueError("Invalid NoneType")
        return default
    if isinstance(t, str):
        t = locate(t)
    if not isinstance(t, type):
        raise TypeError(f"Invalid type: {type(t)}")
    return t


def parse_name(name: str) -> str:
    return " ".join(
        [
            s.upper() if len(s) <= 3 else s.title()
            for s in re.sub(r"[,._\- ]", " ", re.sub(r"[^0-9A-Za-zäöüÄÖÜß%&;:()_.\- ]+", "", name)).split()
        ]
    )


# noinspection PyShadowingBuiltins
def validate_key(id: str) -> str:
    for c in INVALID_CHARS:
        id = id.replace(c, "_")
    return re.sub(r"\W", "", id).lower()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
lories
~~~~~~

To learn how to use local resource integration systems, see "lories --help"

"""

import os
from argparse import ArgumentParser, RawTextHelpFormatter

os.environ["NUMEXPR_MAX_THREADS"] = str(os.cpu_count())


def main() -> None:
    import lories

    parser = ArgumentParser(description=__doc__, formatter_class=RawTextHelpFormatter)
    parser.add_argument("-v", "--version", action="version", version=f"%(prog)s {lories.__version__}")

    application = lories.load(parser=parser)
    application.main()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
lories.connectors.tasks.task
~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from threading import Thread
from typing import Any

from lories._core._channel import ChannelState  # noqa
from lories._core._channels import Channels  # noqa
from lories._core._connector import Connector  # noqa
from lories.connectors.errors import ConnectionError, ConnectorError


class ConnectorTask(ABC, Thread):
    connector: Connector
    channels: Channels

    def __init__(self, connector: Connector, channels: Channels, name: str = None, **kwargs):
        super().__init__(name=name, target=self.__call__, **kwargs)
        self._logger = logging.getLogger(self.__module__)
        self.connector = connector
        self.channels = channels

    # noinspection PyUnresolvedReferences
    def __call__(self, **kwargs) -> Any:
        try:
            return self.run(**kwargs)

        except ConnectionError as e:
            try:
                self.connector.set_channels(ChannelState.DISCONNECTING)
                self.connector.disconnect()
            finally:
                self.connector.set_channels(ChannelState.DISCONNECTED)
                raise e
        except ConnectorError as e:
            raise e
        except Exception as e:
            raise ConnectorError(self.connector, str(e))

    @abstractmethod
    def run(self, **kwargs) -> Any:
        pass



# -*- coding: utf-8 -*-
"""
lories.connectors.sql.columns.surrogate
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from lories.connectors.sql.columns import Column, ColumnType


# noinspection PyShadowingBuiltins
class SurrogateKeyColumn(Column):
    inherit_cache: bool = True

    def __init__(
        self,
        name: str,
        type: ColumnType,
        attribute: str,
        *args,
        nullable: bool = False,
        primary_key: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(
            name,
            type,
            *args,
            nullable=nullable,
            primary_key=primary_key,
            **kwargs,
        )
        self.attribute = attribute



# -*- coding: utf-8 -*-
"""
lories.core.register._load
~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
import os
from abc import abstractmethod
from copy import deepcopy
from logging import Logger
from typing import Any, Collection, Dict, Generic, Mapping, Optional, Sequence, Type

from lories._core._configurator import Configurator  # noqa
from lories._core._registrator import (  # noqa
    Registrator,
    RegistratorContext,
    _RegistratorContext,  # noqa
)
from lories.core.configs import Configurations, Directories, Directory
from lories.core.errors import ResourceError
from lories.util import update_recursive

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import get_args, get_origin

except ImportError:
    from typing_extensions import get_args, get_origin

# FIXME: Remove this once Python >= 3.12 is a requirement
try:
    from typing import get_bound

except ImportError:

    def get_bound(cls):
        return getattr(cls, "__bound__", None)


# noinspection SpellCheckingInspection, PyAbstractClass, PyProtectedMember
class _RegistratorLoader(_RegistratorContext[Registrator], Generic[Registrator]):
    __registrator_class: Type[Registrator]
    _logger: Logger

    def __init__(
        self,
        logger: Optional[Logger] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        if logger is None:
            logger = self._build_logger()
        self._logger = logger
        self.__registrator_class = self._build_registrators_class()

    @classmethod
    def _build_logger(cls) -> Logger:
        return logging.getLogger(cls.__module__)

    @classmethod
    def _build_registrators_class(cls) -> Type[Registrator]:
        for _cls in cls.mro():
            for _base in getattr(_cls, "__orig_bases__", ()):
                if isinstance(_base, type):
                    continue
                if issubclass(get_origin(_base), _RegistratorContext):
                    (_arg,) = get_args(_base)
                    if not isinstance(_arg, type):
                        _arg = get_bound(_arg)
                    return _arg
        raise ResourceError("Unable to extract context registrator")

    # noinspection PyUnresolvedReferences
    def _build_registrator_id(
        self,
        context: RegistratorContext,
        configs: Configurations,
    ) -> str:
        return self._get_registator_class()._build_id(context=context, configs=configs)

    # noinspection PyUnresolvedReferences
    def _build_registrator_defaults(
        self,
        configs: Configurations,
        includes: Optional[Collection[str]] = (),
    ) -> Dict[str, Any]:
        return self._get_registator_class()._build_defaults(configs, includes)

    # noinspection PyUnresolvedReferences
    def _load(
        self,
        context: RegistratorContext,
        configs: Configurations,
        configs_file: Optional[str] = None,
        configs_dir: Optional[str | Directory] = None,
        includes: Optional[Collection[str]] = (),
        defaults: Optional[Mapping[str, Any]] = None,
        configure: bool = True,
        sort: bool = True,
        **kwargs: Any,
    ) -> Sequence[Registrator]:
        registrators = []
        if defaults is None:
            defaults = {}
        update_recursive(defaults, self._build_registrator_defaults(configs, includes))

        configs_dirs = configs.dirs.copy()
        if configs_dir is None:
            configs_dir = configs.dirs.conf.joinpath(configs.name.replace(".conf", ".d"))
        if isinstance(configs_dir, str) and not os.path.isabs(configs_dir):
            configs_dir = configs_dirs.conf.joinpath(configs_dir)
        configs_dirs.conf = configs_dir
        configs_members = configs.get_members([s for s in configs.members if s not in defaults])

        if configs_file and configs_file != configs.name:
            registrators.extend(self._load_from_file(context, configs_file, configs.dirs, defaults, **kwargs))

        registrators.extend(self._load_from_members(context, configs_members, defaults, **kwargs))
        registrators.extend(self._load_from_dir(context, configs_dirs, defaults, **kwargs))

        if sort:
            self.sort()
        if configure:
            self.configure(registrators)
        return registrators

    def _load_from_configs(
        self,
        context: RegistratorContext,
        configs: Configurations,
        **kwargs: Any,
    ) -> Registrator:
        registrator_id = self._build_registrator_id(context, configs)
        if self._contains(registrator_id):
            registrator = self._update(registrator_id, configs)
        else:
            registrator = self._create(context, configs, **kwargs)
            self._add(registrator)
        return registrator

    # noinspection PyUnresolvedReferences
    def _load_from_members(
        self,
        context: RegistratorContext,
        configs: Configurations,
        includes: Optional[Collection[str]] = (),
        defaults: Optional[Mapping[str, Any]] = None,
        **kwargs: Any,
    ) -> Sequence[Registrator]:
        registrators = []
        if defaults is None:
            defaults = {}
        update_recursive(defaults, self._build_registrator_defaults(configs))

        for member_name in configs.members:
            if member_name in includes:
                continue
            member_file = f"{member_name}.conf"
            member_default = deepcopy(defaults)
            update_recursive(member_default, configs.get(member_name))

            member = Configurations.load(
                member_file,
                **configs.dirs.to_dict(),
                **member_default,
                require=False,
            )
            registrators.append(self._load_from_configs(context, member, **kwargs))
        return registrators

    def _load_from_file(
        self,
        context: RegistratorContext,
        configs_file: str,
        configs_dirs: Directories,
        defaults: Optional[Mapping[str, Any]] = None,
        **kwargs: Any,
    ) -> Sequence[Registrator]:
        registrators = []
        if configs_dirs.conf.joinpath(configs_file).is_file():
            # Do not call .load() function here, as configs_dirs._conf may be None and would otherwise be overridden
            # with the data directory
            configs = Configurations(configs_file, deepcopy(configs_dirs))
            configs._load()
            registrators.extend(self._load_from_members(context, configs, defaults, **kwargs))
        return registrators

    def _load_from_dir(
        self,
        context: RegistratorContext,
        configs_dirs: Directories,
        defaults: Optional[Mapping[str, Any]] = None,
        **kwargs: Any,
    ) -> Sequence[Registrator]:
        registrators = []
        if os.path.isdir(configs_dirs.conf):
            for configs_entry in os.scandir(configs_dirs.conf):
                if (
                    configs_entry.is_file()
                    and configs_entry.path.endswith(".conf")
                    and not configs_entry.path.endswith("default.conf")
                    and configs_entry.name
                    not in [
                        "settings.conf",
                        "system.conf",
                        "evaluations.conf",
                        "replications.conf",
                        "logging.conf",
                    ]
                ):
                    configs = Configurations.load(
                        configs_entry.name,
                        **configs_dirs.to_dict(),
                        **defaults,
                    )
                    try:
                        registrators.append(self._load_from_configs(context, configs, **kwargs))

                    except ResourceError:
                        # Skip files with missing or unknown type
                        # TODO: Introduce debug logging here
                        pass
        return registrators

    # noinspection PyUnresolvedReferences, PyTypeChecker
    def configure(self, configurators: Optional[Collection[Configurator]] = None) -> None:
        if configurators is None:
            configurators = self.values()
        for configurator in configurators:
            configurations = configurator.configs
            if configurations is None or not configurations.enabled:
                self._logger.debug(f"Skipping configuring disabled {type(configurator).__name__}")
                continue

            self._logger.debug(f"Configuring {type(self).__name__}: {configurations.path}")
            configurator.configure(configurations)

            if self._logger.getEffectiveLevel() <= logging.DEBUG:
                self._logger.debug(f"Configured {configurator}")

    # noinspection PyUnresolvedReferences, PyArgumentList, PyShadowingBuiltins
    def _update(self, id: str, configs: Configurations) -> Registrator:
        registrator = self._get(id)
        if registrator.is_configured() and configs.enabled:
            registrator_configs = registrator.configs.copy()
            registrator_configs.update(configs)
            registrator.update(registrator_configs)
        else:
            registrator.configs.update(configs)
        return registrator

    # noinspection PyUnresolvedReferences
    def get_all(self, *types: Type[Registrator]) -> Collection[Registrator]:
        if len(types) == 0:
            return self.values()
        return self.filter(lambda r: any(isinstance(r, t) and r.is_enabled() for t in types))

    def get_first(self, *types: Optional[str | Type[Registrator]]) -> Optional[Registrator]:
        registrators = self.get_all(*types)
        return next(iter(registrators)) if len(registrators) > 0 else None

    # noinspection PyTypeChecker
    def get_last(self, *types: Optional[str | Type[Registrator]]) -> Optional[Registrator]:
        registrators = self.get_all(*types)
        return next(reversed(registrators)) if len(registrators) > 0 else None

    def has_type(self, *types: str | Type[Registrator]) -> bool:
        if len(types) == 0:
            raise ValueError("At least one type to look up required")
        return len(self.get_all(*types)) > 0

    # noinspection PyTypeChecker, PyUnresolvedReferences
    def _get_registator_class(self) -> Type[Registrator]:
        return self.__registrator_class

    @classmethod
    def _get_registrators_type(cls) -> str:
        return cls.TYPE

    @abstractmethod
    def _load_registrators_configs(self) -> Configurations: ...

    @abstractmethod
    def load(self, **kwargs: Any) -> Sequence[Registrator]:
        pass



# -*- coding: utf-8 -*-
"""
lories._core._configurations
~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import datetime as dt
from abc import ABC, abstractmethod
from collections.abc import MutableMapping
from typing import Any, Collection, Iterable, Mapping, Optional, Sequence, TypeVar

import pandas as pd


class _Configurations(ABC, MutableMapping[str, Any]):
    @property
    @abstractmethod
    def key(self) -> str: ...

    @abstractmethod
    def remove(self, *keys: str) -> None: ...

    @abstractmethod
    def pop(self, key: str, default: Any = None) -> Any: ...

    @abstractmethod
    def set(self, key: str, value: Any, replace: bool = True) -> None: ...

    @abstractmethod
    def get(self, key: str | Iterable[str], default: Any = None) -> Any: ...

    @abstractmethod
    def get_bool(self, key: str, default: bool = None) -> bool: ...

    @abstractmethod
    def get_int(self, key: str, default: int = None) -> int: ...

    @abstractmethod
    def get_float(self, key: str, default: float = None) -> float: ...

    def get_date(self, key: str, default: pd.Timestamp | dt.datetime = None, **kwargs) -> pd.Timestamp: ...

    @abstractmethod
    def move_to_top(self, key: str) -> None: ...

    @abstractmethod
    def move_to_bottom(self, key: str) -> None: ...

    @abstractmethod
    def write(self, **kwargs) -> None: ...

    @abstractmethod
    def copy(self, **kwargs) -> Configurations: ...

    @property
    @abstractmethod
    def enabled(self) -> bool: ...

    @property
    @abstractmethod
    def members(self) -> Sequence[str]: ...

    @abstractmethod
    def has_member(self, key: str, includes: bool = False) -> bool: ...

    @abstractmethod
    def get_members(
        self,
        keys: Collection[str],
        ensure_exists: bool = False,
    ) -> Configurations: ...

    @abstractmethod
    def get_member(
        self,
        key: str,
        defaults: Optional[Mapping[str, Any]] = None,
        ensure_exists: bool = False,
    ) -> Configurations: ...

    @abstractmethod
    def pop_member(
        self,
        key: str,
        defaults: Optional[Mapping[str, Any]] = None,
    ) -> Configurations: ...

    @abstractmethod
    def update(self, update: Mapping[str, Any], replace: bool = True) -> None: ...


Configurations = TypeVar("Configurations", bound=_Configurations)



# -*- coding: utf-8 -*-
"""
lories.application.view.pages.group
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import re
from collections.abc import MutableSequence, Sequence
from itertools import chain
from typing import Any, Collection, Generic, Iterator, List, Optional, Tuple, Type, TypeVar

import dash_bootstrap_components as dbc
from dash import html

import pandas as pd
from lories.application.view.pages import Page, PageLayout

P = TypeVar("P", bound=Page)


class PageGroup(Page, MutableSequence[P], Generic[P]):
    _pages: List[P]

    order: int = 1000

    # noinspection PyShadowingBuiltins
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._pages = []

    def __len__(self) -> int:
        return len(self._pages)

    def __iter__(self) -> Iterator[P]:
        return iter(self._pages)

    def __contains__(self, page: str | P) -> bool:
        return self.has_page(page)

    def __getitem__(self, index: int) -> P:
        return self._pages[index]

    def __setitem__(self, index: int, page: P) -> None:
        self._pages[index] = page

    def __delitem__(self, index: int) -> None:
        del self._pages[index]

    def insert(self, index: int, page: P) -> None:
        self._pages.insert(index, page)

    def has_page(self, _page: str | P) -> bool:
        if isinstance(_page, str):
            return len(self.get_page(_page)) > 0
        elif isinstance(_page, Page):
            return _page in self._pages
        raise TypeError(f"Invalid page type '{type(_page)}'")

    def get_page(self, _page: str) -> Optional[P]:
        for page in self._pages:
            if page.id == _page:
                return page
        return None

    def get_pages(self, *_pages: str | Type[P]) -> Collection[P]:
        pages = []
        for _page in _pages:
            for page in self._pages:
                if isinstance(_page, str):
                    if page.id == _page:
                        pages.append(page)
                elif isinstance(_page, type):
                    if isinstance(page, _page):
                        pages.append(page)
                else:
                    raise TypeError(f"Invalid page type '{type(_page)}'")
        return pages

    def sort(self) -> Sequence[Page]:
        def order(page: Page) -> Tuple[Any, ...]:
            elements = re.split(r"[^0-9A-Za-zäöüÄÖÜß]+", page.id)
            elements = list(chain(*[re.findall(r"\D+|\d+", t) for t in elements]))
            elements = [(int(t), 0) if t.isdigit() else (0, t) for t in elements if pd.notna(t) and t.strip()]
            return page.order, *elements

        self._pages.sort(key=lambda p: order(p))
        return self._pages

    def create_layout(self, layout: PageLayout) -> None:
        layout.container.class_name = "card-container"

        layout.menu = dbc.NavItem(dbc.NavLink(self.name, href=self.path))
        layout.card.add_title(self.name)
        layout.card.add_footer(href=self.path)

        layout.append(dbc.Row(dbc.Col(html.H4(f"{self.name}:"))))

        for page in self.sort():
            if page.layout.has_card_items():
                layout.append(dbc.Row(dbc.Col(page.layout.card, width="auto"), align="stretch"))



# -*- coding: utf-8 -*-
"""
lories.io.csv
~~~~~~~~~~~~~


"""

from __future__ import annotations

import glob
import os
from typing import List, Mapping, Optional

import pandas as pd
import pytz as tz
from lories.core.errors import ResourceError
from lories.typing import Timestamp, Timezone
from lories.util import ceil_date, floor_date, to_date, to_timedelta


# noinspection PyShadowingBuiltins
def has_range(
    path: str,
    freq: str,
    format: str,
    start: Timestamp | str,
    end: Timestamp | str,
    timezone: tz.tzinfo = tz.UTC,
):
    files = get_files(path, freq, format, start, end, timezone)
    return len(files) > 0


# noinspection PyShadowingBuiltins
def read_files(
    path: str,
    freq: str,
    format: str,
    start: Optional[Timestamp | str] = None,
    end: Optional[Timestamp | str] = None,
    timezone: Optional[Timezone] = None,
    **kwargs,
) -> pd.DataFrame:
    data = pd.DataFrame()
    start = to_date(start, timezone)
    end = to_date(end, timezone)

    files = get_files(path, freq, format, start, end, timezone)
    if len(files) == 0:
        return data
    elif len(files) == 1:
        data = read_file(files[0], timezone=timezone, **kwargs)
    else:
        for file in get_files(path, freq, format, start, end, timezone):
            if not data.empty and (end is not None and data.index[-1] > end):
                break
            file_data = read_file(file, timezone=timezone, **kwargs)
            data = pd.concat([data, file_data], axis="index")

    if data.empty:
        return data
    if not pd.isna(start):
        data = data.loc[data.index >= start, :]
    if not pd.isna(end):
        data = data.loc[data.index <= end, :]
    return data


def read_file(
    path: str,
    index_column: str = "Timestamp",
    index_type: str = "Timestamp",
    timezone: Optional[tz.tzinfo] = None,
    separator: str = ",",
    decimal: str = ".",
    rename: Optional[Mapping[str, str]] = None,
    encoding: str = "utf-8-sig",
) -> pd.DataFrame:
    """
    Reads the content of a specified CSV file.

    :param path:
        the full path to the CSV file.
    :type path:
        string

    :param index_column:
        the column name of the CSV file index.
    :type index_column:
        string

    :param index_type:
        the index type, either "Timestamp", "UNIX" oder "None".
    :type index_type:
        string

    :param separator:
        the separator character of the CSV file.
    :type separator:
        string

    :param decimal:
        the decimal character used for the CSV file.
    :type decimal:
        string

    :param timezone:
        the timezone for the timestamp to be converted or localized to.
    :type timezone:
        :class: `pytz.tzinfo`

    :param rename:
        the dictionary to rename columns by after reading, if not None.
    :type rename:
        dict

    :param encoding:
        the encoding to read_file the file with.
    :type encoding:
        string


    :returns:
        the retrieved columns, indexed by their timestamp
    :rtype:
        :class:`pandas.DataFrame`
    """
    data = pd.read_csv(path, sep=separator, decimal=decimal, encoding=encoding)
    if not data.empty:
        if index_column not in data.columns:
            if index_column.islower():
                index_column = index_column.title()
            else:
                index_column = index_column.lower()

        if index_type.lower() in ["timestamp", "unix"]:
            if index_type.lower() == "timestamp":
                data[index_column] = pd.to_datetime(data[index_column], utc=True)
            elif index_type.lower() == "unix":
                data[index_column] = pd.to_datetime(data[index_column], unit="ms")
            else:
                raise ValueError(f"Unknown index type: {index_type}")

            data.set_index(index_column, inplace=True)  # , verify_integrity=True)

            if not hasattr(data.index, "tzinfo"):
                data[index_column] = data.index
                data[index_column] = data[index_column].apply(lambda t: t.astimezone(tz.UTC).replace(tzinfo=None))
                data.set_index(index_column, inplace=True)  # , verify_integrity=True)
                data.index = data.index.tz_localize(tz.UTC)

            if timezone is not None:
                if hasattr(data.index, "tzinfo") and data.index.tzinfo is not None:
                    if data.index.tzinfo != timezone:
                        data.index = data.index.tz_convert(timezone)
                else:
                    data.index = data.index.tz_localize(timezone, ambiguous="infer")

        elif index_type is None or index_type.lower() == "none":
            # Prepare the index name, to be renamed below
            data.index.name = "index"
        else:
            raise ValueError(f"Unknown index type: {index_type}")

    if rename:
        data = data.rename(columns=rename)
        data.index.name = data.index.name.lower()
    else:
        data.index.name = data.index.name.title()
    return data


# noinspection PyShadowingBuiltins
def write_files(
    data: pd.DataFrame,
    path: str,
    freq: str,
    format: str,
    timezone: Optional[tz.tzinfo] = None,
    **kwargs,
) -> None:
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)

    index_name = data.index.name
    if index_name is None:
        index_name = "timestamp"
    if data.index.tzinfo is None or data.index.tzinfo.utcoffset(data.index) is None:
        data.index = data.index.tz_localize(tz.UTC, ambiguous="infer")
    if timezone is None:
        timezone = data.index.tzinfo
    elif data.index.tzinfo != timezone:
        data.index = data.index.tz_convert(timezone)

    time_step = floor_date(data.index[0], freq=freq)

    def next_step() -> pd.Timestamp:
        return floor_date(time_step + to_timedelta(freq), timezone=timezone, freq=freq)

    while time_step < data.index[-1]:
        time_next = next_step()

        file = time_step.strftime(format) + ".csv"
        file_path = os.path.join(path, file)
        file_data = data[(data.index >= time_step) & (data.index < time_next)].copy()
        file_data.index.name = index_name

        write_file(file_data, file_path, timezone=timezone, **kwargs)

        time_step = time_next


def write_file(
    data: pd.DataFrame,
    path: str,
    timezone: Optional[tz.tzinfo] = None,
    separator: str = ",",
    decimal: str = ".",
    rename: Optional[Mapping[str, str]] = None,
    override: bool = False,
    encoding: str = "utf-8-sig",
):
    if data.index.tzinfo is None or data.index.tzinfo.utcoffset(data.index) is None:
        data.index = data.index.tz_localize(timezone, ambiguous="infer")
    if timezone is None:
        timezone = data.index.tzinfo
    elif data.index.tzinfo != timezone:
        data.index = data.index.tz_convert(timezone)

    if not override and os.path.isfile(path):
        index = data.index.name
        csv = read_file(
            path,
            index_column=index,
            timezone=timezone,
            separator=separator,
            decimal=decimal,
            rename={column: name for name, column in rename.items()} if not None else None,
            encoding=encoding,
        )

        if not csv.empty:
            if all(name in list(csv.columns) for name in list(data.columns)):
                data = data.combine_first(csv)
            else:
                data = pd.concat([csv, data], axis="index")

    if rename:
        data = data.rename(columns=rename)
    if data.index.name is None:
        data.index.name = "timestamp"

    data.to_csv(path, sep=separator, decimal=decimal, encoding=encoding)


# noinspection PyShadowingBuiltins
def get_files(
    path: str,
    freq: str,
    format: str,
    start: Optional[Timestamp | str] = None,
    end: Optional[Timestamp | str] = None,
    timezone: tz.tzinfo = tz.UTC,
    exists_only: bool = True,
) -> List[str]:
    end = to_date(end, timezone)
    start = to_date(start, timezone)
    if start is None or end is None:
        filenames = [os.path.basename(f) for f in glob.glob(os.path.join(path, "*.csv"))]
        if len(filenames) > 0:
            filenames.sort()

            def _validate(filename: str) -> bool:
                try:
                    _date = to_date(filename, timezone=timezone, format=f"{format}.csv")
                    return True
                except ValueError:
                    return False

            if start is None and end is None:
                return [os.path.join(path, f) for f in filenames if _validate(f)]

            if start is None:
                start_str = filenames[0].replace(".csv", "")
                start = to_date(start_str, timezone=timezone, format=f"{format}.csv")

                if end is None:
                    end_str = filenames[-1].replace(".csv", "")
                    end = to_date(end_str, timezone=timezone, format=f"{format}.csv")
                    end = ceil_date(end, timezone=timezone, freq=freq)

    date = floor_date(start, timezone=timezone, freq=freq)

    # noinspection PyShadowingNames
    def next_date() -> pd.Timestamp:
        next_date = floor_date(date + to_timedelta(freq), timezone=timezone, freq=freq)
        if next_date == date:
            next_date += to_timedelta(freq)
            next_offset = date.utcoffset() - next_date.utcoffset()
            if next_offset.seconds > 0:
                next_date = floor_date(next_date + next_offset, timezone=timezone, freq=freq)
            else:
                ResourceError(f"Unable to increment date for freq '{freq}'")
        return next_date

    files = []
    file = date.strftime(format) + ".csv"
    file_path = os.path.join(path, file)
    if os.path.isfile(file_path) or not exists_only:
        files.append(file_path)
    if end is not None:
        date = next_date()
        while date <= end:
            file = date.strftime(format) + ".csv"
            file_path = os.path.join(path, file)
            if os.path.isfile(file_path) or not exists_only:
                files.append(file_path)
            date = next_date()

    # TODO: Implement or validate if custom sorting by file format is necessary
    files.sort()

    return files



# -*- coding: utf-8 -*-
"""
lories.connectors.revpi
~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Dict, Optional

from revpimodio2 import EventCallback, RevPiModIO, io

import pandas as pd
import pytz as tz
from lories.connectors import Connector, register_connector_type
from lories.data.channels import Channel
from lories.typing import Configurations, Resources
from lories.util import to_bool


# noinspection PyShadowingBuiltins, SpellCheckingInspection
@register_connector_type("revpi", "revpi_io", "revpi_aio", "revpi_mio", "revpi_ro", "revolutionpi")
class RevPiConnector(Connector):
    _core: RevPiModIO
    _cycletime: Optional[int]

    _listeners: Dict[str, RevPiListener]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._listeners = {}

    def configure(self, configs: Configurations) -> None:
        super().configure(configs)
        self._cycletime = configs.get_int("cycletime", default=None)

    def connect(self, resources: Resources) -> None:
        super().connect(resources)
        self._core = RevPiModIO(autorefresh=True)
        if self._cycletime:
            self._core.cycletime = self._cycletime

        channels = resources.filter(lambda r: isinstance(r, Channel) and to_bool(r.get("listener", False)))
        for channel in channels:
            channel_listener = RevPiListener(channel)
            channel_io = self._core.io[channel_listener.address]
            channel_io.reg_event(channel_listener, edge=io.RISING, as_thread=True, prefire=True)
            self._listeners[channel.id] = channel_listener

        # Handle SIGINT / SIGTERM to exit program cleanly
        # self._core.handlesignalend(self._core.cleanup)

        # TODO: set all IO output values to optional default attribute value
        self._core.mainloop(blocking=False)

    def disconnect(self) -> None:
        super().disconnect()
        for listener in self._listeners.values():
            listener_io = self._core.io[listener.address]
            listener_io.unreg_event(listener)

        # TODO: set all IO output values to optional default attribute value

        self._core.cleanup()

    # noinspection PyTypeChecker
    def read(self, resources: Resources) -> pd.DataFrame:
        now = pd.Timestamp.now(tz=tz.UTC).floor(freq="s")
        data = pd.DataFrame(index=[now], columns=resources.ids)
        for resource in resources:
            resource_io = self._core.io[resource.address]
            self._logger.debug(f"Read RevPi IO '{resource_io}': {resource_io.value}")

            data.at[now, resource.id] = resource_io.value
        return data

    def write(self, data: pd.DataFrame) -> None:
        for channel in self.channels:
            if channel.id not in data.columns:
                continue
            channel_data = data.loc[:, channel.id].dropna(axis="index", how="all")
            if channel_data.empty:
                continue

            channel_io = self._core.io[channel.address]
            channel_io.value = channel_data.iloc[-1]


class RevPiListener:
    address: str

    _channel: Channel

    def __init__(self, channel: Channel):
        self._channel = channel
        self.address = channel.address

    def __call__(self, event: EventCallback) -> None:
        now = pd.Timestamp.now(tz=tz.UTC).floor(freq="s")
        self._channel.set(now, event.iovalue)



# -*- coding: utf-8 -*-
"""
lories.core.constant
~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Any, Dict, Optional, Type

from lories._core import _Constant  # noqa
from lories.core.constants import Constants
from lories.core.errors import ResourceError
from lories.util import parse_type


# noinspection PyPep8Naming, PyShadowingBuiltins
class Constant(_Constant):
    # noinspection PyTypeChecker
    def __new__(cls, type: Type, key: str, name: Optional[str] = None, unit: Optional[str] = None):
        if key in CONSTANTS:
            raise ResourceError(f"Constant '{key}' already exists.")
        constant = str.__new__(cls, key)
        CONSTANTS.append(constant)
        return constant

    def __init__(self, type: Type | str, key: str, name: Optional[str] = None, unit: Optional[str] = None) -> None:
        self.__type = parse_type(type)
        self.__key = key
        self.__name = name
        self.__unit = unit

    def __str__(self) -> str:
        return self.key

    @property
    def type(self) -> Type:
        return self.__type

    @property
    def key(self) -> str:
        return self.__key

    @property
    def name(self) -> str:
        return self.__name

    @property
    def unit(self) -> Optional[str]:
        return self.__unit

    def full_name(self, unit: bool = False) -> str:
        name = self.name
        if unit and self.unit is not None and len(self.unit) > 0:
            name += f" [{self.unit}]"
        return name

    def to_dict(self) -> Dict[str, Any]:
        dict = {
            "type": self.__type,
            "key": self.__key,
            "name": self.__name if self.__name is not None else self.__key.replace("_", " ").title(),
        }
        if self.__unit is not None:
            dict["unit"] = self.__unit
        return dict


CONSTANTS = Constants()

SECOND = Constant(int, "second", "Second")
MINUTE = Constant(int, "minute", "Minute")
HOUR = Constant(int, "hour", "Hour")
DAY_OF_WEEK = Constant(int, "day_of_week", "Day of the Week")
DAY_OF_YEAR = Constant(int, "day_of_year", "Day of the Year")
MONTH = Constant(int, "month", "Month")
YEAR = Constant(int, "year", "Year")



# -*- coding: utf-8 -*-
"""
lories.connectors.sql.column
~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import builtins
import datetime as dt
from typing import Any, AnyStr, Optional, Type, TypeVar

import sqlalchemy as sql
from sqlalchemy.types import BLOB, BOOLEAN, DATETIME, FLOAT, INTEGER, TIMESTAMP, String, TypeEngine

import numpy as np
import pandas as pd
from lories.core import ConfigurationError, ResourceError

ColumnType = TypeVar("ColumnType", Type[TypeEngine], TypeEngine)


class Column(sql.Column):
    inherit_cache: bool = True

    # noinspection PyShadowingBuiltins, SpellCheckingInspection
    def __init__(
        self,
        name: str,
        type: ColumnType,
        *args,
        nullable: bool = True,
        default: Optional[Any] = None,
        onupdate: Optional[Any] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            name,
            type,
            *args,
            nullable=nullable,
            server_default=default,
            server_onupdate=onupdate,
            **kwargs,
        )

    @property
    def _constructor(self):
        # TODO: Look into
        return sql.Column

    def validate(self, data: Any) -> Any:
        if data is None and not self.nullable:
            raise ResourceError(f"None value for '{self.name}' NOT NULL")
        return data


# noinspection PyShadowingBuiltins
def parse_type(type: Type | AnyStr, length: Optional[int] = None) -> Type[TypeEngine] | TypeEngine:
    if isinstance(type, builtins.type):
        if issubclass(type, (pd.Timestamp, dt.datetime, pd.arrays.DatetimeArray)):
            type = "TIMESTAMP"
        elif issubclass(type, (bool, pd.arrays.BooleanArray)):
            type = "BOOL"
        elif issubclass(type, (int, np.integer, pd.arrays.IntegerArray)):
            type = "INT"
        elif issubclass(type, (float, np.floating, pd.arrays.FloatingArray)):
            type = "FLOAT"
        elif issubclass(type, (str, pd.arrays.StringArray)):
            type = "STRING"
        elif issubclass(type, (bytes, bytearray)):
            type = "BYTE"

    if isinstance(type, str):
        type = type.upper()
    else:
        raise ConfigurationError(f"Unknown SQL data type: {type}")

    return to_type_engine(type, length)


# noinspection PyShadowingBuiltins
def to_type_engine(type: Type | AnyStr, length: Optional[int] = None) -> Type[TypeEngine] | TypeEngine:
    if type == "FLOAT":
        return FLOAT
    if type in ["INT", "INTEGER"]:
        return INTEGER
    if type in ["BOOL", "BOOLEAN"]:
        return BOOLEAN
    if type in ["BYTE", "BYTES"]:
        return BLOB(length=4294967295)
    if type == "DATETIME":
        return DATETIME
    if type == "TIMESTAMP":
        return TIMESTAMP
    if type in ["VARCHAR", "STRING"]:
        if type == "VARCHAR" and length is None:
            raise ConfigurationError(f"Invalid SQL data type '{type}' without configured length")
        if length is None:
            return String
        return String(length)

    raise ConfigurationError(f"Unknown SQL data type: {type}")



# -*- coding: utf-8 -*-
"""
lories.simulation.results
~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import os
import re
from collections.abc import Callable
from functools import reduce
from typing import Any, Dict, Iterable, Iterator, List, Optional, Sequence

import pandas as pd
from lories.components import Component
from lories.connectors import Database
from lories.connectors.tables import HDFDatabase
from lories.core import CONSTANTS, Configurator, Constant, Directories, ResourceError, Resources
from lories.core.typing import Configurations, Timestamp
from lories.data.util import resample, scale_energy, scale_power
from lories.simulation import Durations, Progress, Result
from lories.util import parse_freq


class Results(Configurator, Sequence[Result]):
    __list: List[Result]

    __resources: Resources
    __component: Component
    __database: Database

    _freq: Optional[str]

    data: pd.DataFrame

    dirs: Directories

    durations: Durations
    progress: Progress

    def __init__(
        self,
        component: Component,
        database: Database,
        configs: Configurations,
        desc: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(configs)
        if database is None:
            database_configs = component.configs.get_member(
                key="connectors",
                ensure_exists=True,
            ).get_member(
                key="results",
                defaults={},
            )
            database_configs.update(
                {
                    "key": "results",
                    "type": "tables",
                    "file": str(component.configs.dirs.data.joinpath(".results.h5")),
                    "compression_level": 9,
                    "compression_lib": "zlib",
                }
            )
            database = HDFDatabase(component, configs=database_configs)
            database.configure(database_configs)
            component.connectors.add(database)

        self.__list = []
        self.__database = self._assert_database(database)
        self.__component = self._assert_component(component)
        self.__resources = self._extract_resources(component)

        self.dirs = component.configs.dirs
        if not self.dirs.data.exists():
            os.makedirs(self.dirs.data, exist_ok=True)
        self.dirs.tmp = self.dirs.data.joinpath(".results")
        if not self.dirs.tmp.exists():
            os.makedirs(self.dirs.tmp, exist_ok=True)

        if desc is None:
            desc = component.name
        self.progress = Progress(
            desc=desc,
            file=str(self.dirs.data.joinpath("results.json")),
            **kwargs,
        )
        self.durations = Durations(self.dirs.tmp)
        self.data = pd.DataFrame()

    @classmethod
    def _assert_database(cls, database: Database) -> Database:
        if database is None or not isinstance(database, Database):
            raise ResourceError(f"Invalid '{cls.__name__}' database: {type(database)}")
        return database

    @classmethod
    def _assert_component(cls, component: Component) -> Component:
        if component is None or not isinstance(component, Component):
            raise ResourceError(f"Invalid '{cls.__name__}' component: {type(component)}")
        return component

    @staticmethod
    def _extract_resources(component: Component) -> Resources:
        resources = []

        def extend_resources(__component: Component) -> None:
            resources.extend(__component.data.channels)
            for _component in __component.components.values():
                extend_resources(_component)

        extend_resources(component)
        return Resources(resources)

    def configure(self, configs: Configurations) -> None:
        super().configure(configs)
        self._freq = parse_freq(configs.get("freq", default=None))

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join(str(r.key) for r in self.__list)})"

    def __str__(self) -> str:
        return f"{type(self).__name__}:\n\t" + "\n\t".join(f"{r.key} = {repr(r)}" for r in self.__list)

    def __contains__(self, result: str | Result) -> bool:
        if isinstance(result, str):
            return any(result == r.key for r in self.__list)
        return result in self.__list

    def __getitem__(self, index: Iterable[str] | str | int):
        if isinstance(index, str):
            for result in self.__list:
                if result.key == index:
                    return result
        if isinstance(index, Iterable):
            return [r for r in self.__list if r.key == index]
        raise KeyError(index)

    def __iter__(self) -> Iterator[Result]:
        return iter(self.__list)

    def __len__(self) -> int:
        return len(self.__list)

    def __add__(self, other):
        return [*self, *other]

    def add(
        self,
        key: str | Constant,
        name: str,
        summary: Any,
        header: str = "Summary",
        **kwargs: Any,
    ) -> None:
        self.__list.append(Result(key, name, summary, header=header, **kwargs))

    def append(self, result: Result) -> None:
        self.__list.append(result)

    def extend(self, results: Iterable[Result]) -> None:
        self.__list.extend(results)

    def __enter__(self, **kwargs) -> Results:
        configs = self.configs
        self.configure(configs)
        self.open()
        return self

    # noinspection PyShadowingBuiltins
    def __exit__(self, type, value, traceback):
        self.close()

    def open(self) -> None:
        self.__database.connect(self.__resources)

    def close(self) -> None:
        self.__database.disconnect()
        self.durations.stop()
        self.progress.close()

    @property
    def key(self) -> str:
        return self.__component.key

    @property
    def name(self) -> str:
        return self.__component.name

    @property
    def start(self) -> Optional[pd.Timestamp]:
        if len(self.data.index) == 0:
            return None
        return self.data.index[0]

    @property
    def end(self) -> Optional[pd.Timestamp]:
        if len(self.data.index) < 2:
            return None
        return self.data.index[-1]

    # noinspection PyShadowingBuiltins
    def filter(self, filter: Callable[[Result], bool]) -> Sequence[Result]:
        return [result for result in self.__list if filter(result)]

    def report(self) -> None:
        self.durations.complete()
        self.progress.complete(**self.to_dict())
        try:
            # TODO: Make reports configurable
            from lories.io import excel

            excel_file = str(self.dirs.data.joinpath("results.xlsx"))
            excel.write(excel_file, "Results", self.to_frame())

            if self.configs.get_bool("include", default=True):
                columns = {c.key: c.full_name(unit=True) for c in CONSTANTS}
                columns.update({r.get("column", default=r.key): r.full_name(unit=True) for r in self.__resources})

                if self._freq is not None:
                    resampled = []
                    for method, resources in self.__resources.groupby("aggregate"):
                        resample_columns = [r.get("column", default=r.key) for r in resources]
                        resample_columns = [c for c in resample_columns if c in self.data.columns]
                        if len(resample_columns) == 0:
                            continue
                        if method is None:
                            self._logger.warning(
                                "Skipping resources for missing aggregate function: "
                                + ", ".join(f"'{r.id}'" for r in resources)
                            )
                            continue
                        resampled.append(resample(self.data[resample_columns], self._freq, method))

                    if len(resampled) == 0:
                        data = pd.DataFrame()
                    else:
                        data = pd.concat(resampled, axis="columns")[self.data.columns]
                        data.rename(inplace=True, columns=columns)
                else:
                    data = self.data.rename(columns=columns)
                excel.write(excel_file, "Timeseries", data)

        except ImportError:
            pass

    # noinspection PyTypeChecker
    def submit(
        self,
        function: Callable[..., pd.DataFrame],
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        *args,
        **kwargs,
    ) -> None:
        columns = {r.get("column", default=r.key): r.id for r in self.__resources}

        if self.__database.exists(self.__resources, start, end):
            data = self.__database.read(self.__resources, start, end)
            data.rename(columns={v: k for k, v in columns.items()}, inplace=True)
        else:
            data = function(start, end, *args, **kwargs)
            self.__database.write(data.rename(columns=columns))

        self.data = pd.concat([self.data, data], axis="index")

    def is_complete(self) -> bool:
        return len(self.__list) > 0 and self.durations.is_complete()

    def is_success(self) -> bool:
        return self.is_complete() and self.progress.status == "success"

    def to_dict(self) -> Dict[str, Any]:
        return {r.key: r.summary for r in self.__list}

    def to_frame(self) -> pd.DataFrame:
        summary = pd.DataFrame(columns=pd.MultiIndex.from_tuples((), names=["System", ""]))
        order = list(dict.fromkeys(r.order for r in self.__list))
        order.sort()
        for result in reduce(lambda r1, r2: r1 + r2, [self.filter(lambda r: r.order == o) for o in order]):
            name = result.name
            value = result.summary
            if re.search(r".*\[.*kWh.*]", name):
                name, value = scale_energy(name, value)
            elif re.search(r".*\[.*W.*]", name):
                name, value = scale_power(name, value)
            summary.loc[self.name, (result.header, name)] = value
        return summary



# -*- coding: utf-8 -*-
"""
lories.application.view.pages.components.weather
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import List

import dash_bootstrap_components as dbc
import plotly.graph_objs as go
from dash import Input, Output, callback, dcc, html

import pandas as pd
import pytz as tz
from lories.application.view.pages import ComponentPage, PageLayout, register_component_group, register_component_page
from lories.components.weather import WeatherForecast as Forecast
from lories.components.weather import WeatherProvider as Weather
from lories.util import floor_date


@register_component_page(Weather)
@register_component_group(Weather, name="Weather")
class WeatherPage(ComponentPage[Weather]):
    # noinspection PyProtectedMember
    @property
    def forecast(self) -> Forecast:
        return self._component.forecast

    def create_layout(self, layout: PageLayout) -> None:
        super().create_layout(layout)

        current = self._build_current()
        layout.card.append(current, focus=True)
        layout.append(dbc.Row(dbc.Col(current, width="auto")))

        if self.forecast.is_enabled():
            forecast = self._build_forecast()
            layout.card.append(forecast)
            layout.append(dbc.Row(dbc.Col(forecast, width="auto")))

    def _build_current(self) -> html.Div:
        @callback(
            Output(f"{self.id}-current", "children"),
            Input(f"{self.id}-current-update", "n_intervals"),
        )
        def _update_current(*_) -> List[html.P] | dbc.Spinner:
            icon = self.data.icon
            if icon.is_valid():
                return [
                    html.P(icon.timestamp.strftime("Today, %d. %b %Y from %H:%M")),
                    html.P(str(icon.value).replace("-", " "), style={"fontSize": "4rem"}),
                ]
            return dbc.Spinner(html.Div(id=f"{self.id}-current-loader"))

        return html.Div(
            [
                html.Div(
                    _update_current(),
                    id=f"{self.id}-current",
                ),
                dcc.Interval(
                    id=f"{self.id}-current-update",
                    interval=60000,
                    n_intervals=0,
                ),
            ]
        )

    def _build_forecast(self) -> html.Div:
        @callback(
            Output(f"{self.id}-forecast-graph", "figure"),
            Output(f"{self.id}-forecast-icons", "children"),
            Input(f"{self.id}-forecast-update", "n_intervals"),
        )
        def _update_forecast(*_):
            start = floor_date(pd.Timestamp.now(tz.UTC), freq="h")
            end = start + pd.Timedelta(hours=23)
            forecast = self.forecast.get(start, end)
            if forecast.empty:
                forecast = pd.DataFrame(index=pd.date_range(start, end, freq="h"), columns=["icon", Weather.TEMP_AIR])

            return self._build_forecast_graph(forecast), self._build_forecast_icons(forecast)

        return html.Div(
            [
                dbc.Row(id=f"{self.id}-forecast-icons"),
                dbc.Row(
                    dbc.Col(
                        [
                            dcc.Graph(id=f"{self.id}-forecast-graph", config={"displayModeBar": False}),
                            dcc.Interval(id=f"{self.id}-forecast-update", interval=60000, n_intervals=0),
                        ]
                    )
                ),
            ],
            id=f"{self.id}-forecast",
            # className="weather-forecast",
            # TODO: Move this to separate style sheet
            style={
                "padding-top": "1rem",
                "padding-bottom": "1rem",
                "background-image": "linear-gradient(#91d0ff, #c2e4ff)",
                "background-repeat": "no-repeat",
                # "border-radius": "var(--bs-border-radius)",
                # Background should stop above padding x axis margin (32px)
                "background-size": "100% calc(100% - 32px - 1rem)",
            },
        )

    def _build_forecast_graph(self, forecast: pd.DataFrame):
        temperature = forecast[Weather.TEMP_AIR]

        # Create the graph trace
        data = go.Scatter(
            x=temperature.index,
            y=temperature.values,
            mode="lines",
            line={"color": "#e4002c"},
        )

        ticks = self._slice_forecast_index(forecast)

        # Create the graph layout
        layout = go.Layout(
            xaxis=go.layout.XAxis(
                visible=True,
                range=[min(temperature.index), max(temperature.index)],
                tickmode="array",
                tickvals=ticks,
                ticktext=[t.strftime("%H:%M") for t in ticks],
            ),
            yaxis=go.layout.YAxis(
                visible=False,
                range=[
                    round(min(temperature.values) - 10),
                    round(max(temperature.values) + 10),
                ],
            ),
            margin=go.layout.Margin(l=0, r=0, b=32, t=12),
            height=250,
            dragmode=False,
            paper_bgcolor="rgba(0,0,0,0)",
            plot_bgcolor="rgba(0,0,0,0)",
        )
        return {"data": [data], "layout": layout}

    def _build_forecast_icons(self, forecast: pd.DataFrame):
        icons = []
        for timestamp in self._slice_forecast_index(forecast):
            icon = forecast.loc[timestamp, "icon"]
            if not pd.isna(icon):
                icons.append(dbc.Col(html.Div(icon, style={"fontSize": ".75rem", "text-align": "center"})))
            else:
                icons.append(dbc.Col(dbc.Spinner()))
        return icons

    @staticmethod
    def _slice_forecast_index(forecast: pd.DataFrame, hours: int = 4) -> pd.DatetimeIndex:
        times = []
        time = floor_date(forecast.index[0], freq=f"{hours}h")
        while time <= forecast.index[-1]:
            if time in forecast.index:
                times.append(time)
            time += pd.Timedelta(hours=hours)
        return pd.DatetimeIndex(times)



# -*- coding: utf-8 -*-
"""
lories.application.view.pages.components.group
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Collection, Dict, Generic, Optional, Type, TypeVar

import dash_bootstrap_components as dbc
from dash import html

from lories.application.view.pages import PageGroup, PageLayout
from lories.application.view.pages.components import ComponentPage
from lories.components import Component
from lories.core.typing import Channels

ComponentType = TypeVar("ComponentType", bound=Component)
ChildrenType = TypeVar("ChildrenType", bound=Dict[str, Type[ComponentPage]])


# noinspection PyShadowingBuiltins
class ComponentGroup(PageGroup[ComponentPage], ComponentPage[ComponentType], Generic[ComponentType]):
    def __init__(self, component: ComponentType, *args, **kwargs) -> None:
        super().__init__(component=component, *args, **kwargs)

    def _at_create_layout(self, layout: PageLayout) -> None:
        super()._at_create_layout(layout)
        for page in self:
            page.create_layout(page.layout)

    def _create_data_layout(self, layout: PageLayout, channels: Channels, title: Optional[str] = "Data") -> None:
        if len(channels) > 0:
            data = []
            if title is not None:
                data.append(html.H5(f"{title}:"))
            data.append(self._build_data(channels))
            layout.append(dbc.Row(dbc.Col(dbc.Card(dbc.CardBody(data)))))

    def get_page(self, _page: str | Component) -> Optional[ComponentPage]:
        if isinstance(_page, Component):
            _page = f"{self.id}-{self._encode_id(_page.key)}"
        elif not isinstance(_page, str):
            raise TypeError(f"Invalid page type '{type(_page)}'")
        return super().get_page(_page)

    def get_pages(self, *_pages: str | Type[Component | ComponentPage]) -> Collection[ComponentPage]:
        pages = []
        for _page in _pages:
            for page in self._pages:
                if isinstance(_page, str):
                    if page.id == _page:
                        pages.append(page)
                elif isinstance(_page, type):
                    if isinstance(page, _page) or isinstance(page._component, _page):
                        pages.append(page)
                else:
                    raise TypeError(f"Invalid page type '{type(_page)}'")
        return pages



# -*- coding: utf-8 -*-
"""
lories.connectors.serial.sdi12
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import time
from typing import AnyStr, Dict, Optional

import pandas as pd
import pytz as tz
from lories.connectors import register_connector_type
from lories.connectors.serial._serial import _SerialConnector
from lories.data import ChannelState
from lories.typing import Resources
from lories.util import is_int


@register_connector_type("sdi12")
class Sdi12Connector(_SerialConnector):
    def read(self, resources: Resources) -> pd.DataFrame:
        """Read all sensors."""
        timestamp = pd.Timestamp.now(tz=tz.UTC).floor(freq="s")

        results = {}
        for sensor_address, sensor_resources in resources.groupby("sensor"):
            if not is_int(sensor_address):
                self._logger.warning(f"Invalid SDI12 sensor address: {sensor_address}")
                continue
            sensor_data = self._read_sensor(sensor_resources, str(sensor_address))
            if sensor_data is not None:
                results.update(sensor_data)

        if len(results) == 0:
            return pd.DataFrame()
        return pd.DataFrame(index=[timestamp], data=results)

    def _read_sensor(
        self,
        resources: Resources,
        address: AnyStr,
    ) -> Optional[Dict[str, float]]:
        """
        Perform a full measurement cycle:
        aM! → parse atttn → wait ttt → aD0! … aDn! → parse values
        """
        self._break()
        self._write_string(f"{address}M!\r\n")
        response = self._read_line()
        if not response.startswith(address):
            self._logger.warning(f"Invalid SDI12 response to M!: {response}")
            return None

        # Extract the time to wait in seconds
        ttt = int(response[len(address):len(address) + 3])
        if ttt > 0:
            time.sleep(ttt)

        results = {}
        for data, data_resources in resources.groupby("data"):
            self._break()
            self._write_string(f"{address}D{data}!\r\n")

            response = self._read_line()
            if not response or not response.startswith(address):
                self._logger.warning(f"Invalid SDI12 response to D{data}!: {response}")
                break

            response_parts = response[1:].replace("+", " +").replace("-", " -").split()
            for index, index_resources in data_resources.groupby("index"):
                if index is None:
                    index = 0
                elif not is_int(index):
                    self._logger.warning(f"Invalid SDI12 data index: {index}")
                    continue
                index = int(index)
                
                for resource in index_resources:
                    try:
                        results[resource.id] = float(response_parts[index])

                    except (IndexError, ValueError):
                        results[resource.id] = ChannelState.READ_ERROR
                        self._logger.warning(
                            f"Failed to parse SDI12 value for sensor {address} from response: {response}"
                        )
        if len(results) == 0:
            return None
        return results

    def _break(self) -> None:
        """Issue SDI-12 break (≥12 ms of spacing, i.e. logic 0)."""
        self._serial.break_condition = True
        time.sleep(0.015)
        self._serial.break_condition = False

        # ≥8.33 ms mark before next command
        time.sleep(0.0085)

    def write(self, data: pd.DataFrame) -> None:
        raise NotImplementedError("SDI12 does not support writing data")



# -*- coding: utf-8 -*-
"""
lories.connectors.cameras
~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from .camera import CameraConnector  # noqa: F401

from . import opencv  # noqa: F401
from .opencv import OpenCV  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories.application.view.login
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from . import user  # noqa: F401
from .user import User  # noqa: F401

from . import authentication  # noqa: F401
from .authentication import (  # noqa: F401
    Authentication,
    AuthenticationException,
)

from . import page  # noqa: F401
from .page import LoginPage  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories._core._converter
~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from abc import abstractmethod
from typing import Any, Collection, Generic, Optional, Type, TypeAlias, TypeVar, overload

import pandas as pd
from lories._core._channel import Channel
from lories._core._channels import Channels
from lories._core._registrator import _Registrator, _RegistratorContext
from lories._core.typing import Timestamp

T = TypeVar("T", bound=Any)


class _Converter(_Registrator, Generic[T]):
    INCLUDES: Collection[str] = ()
    TYPE: str = "converter"

    @property
    @abstractmethod
    def dtype(self) -> Type[T]: ...

    @abstractmethod
    def is_dtype(self, value: Any) -> bool: ...

    @abstractmethod
    def to_dtype(self, value: Any, **kwargs) -> Optional[T]: ...

    @overload
    def to_str(self, value: T) -> str: ...

    @overload
    def to_str(self, value: pd.Series) -> str: ...

    def to_str(self, value: T | pd.Series) -> str:
        return self.to_json(value)

    @overload
    def to_json(self, value: T) -> str: ...

    @overload
    def to_json(self, value: pd.Series) -> str: ...

    @overload
    def to_json(self, value: pd.DataFrame) -> str: ...

    @abstractmethod
    def to_json(self, data: T | pd.Series | pd.DataFrame) -> str: ...

    @abstractmethod
    def to_series(
        self,
        value: T,
        timestamp: Optional[Timestamp] = None,
        name: Optional[str] = None,
    ) -> pd.Series: ...

    @abstractmethod
    def from_frame(
        self,
        data: pd.DataFrame,
        channels: Channels,
    ) -> pd.DataFrame: ...

    @abstractmethod
    def from_series(
        self,
        data: pd.Series,
        channel: Channel,
    ) -> pd.Series: ...

    @abstractmethod
    def convert(self, value: Any, **kwargs) -> Optional[T]: ...


Converter = TypeVar("Converter", bound=_Converter)


# noinspection PyAbstractClass
class _ConverterContext(_RegistratorContext[Converter]):
    TYPE: str = "converters"


ConverterContext = TypeVar(
    name="ConverterContext",
    bound=_ConverterContext,
)
Converters: TypeAlias = ConverterContext



# -*- coding: utf-8 -*-
"""
lories.core.location
~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Optional

import pytz as tz
from lories.core import ResourceError, ResourceUnavailableError  # noqa
from lories.util import to_timezone


class Location:
    """
    Location objects are convenient containers for latitude, longitude,
    timezone, and altitude data associated with a particular
    geographic location.

    Parameters
    ----------
    latitude : float.
        Positive is north of the equator.
        Use decimal degrees notation.
    longitude : float.
        Positive is east of the prime meridian.
        Use decimal degrees notation.
    timezone : str or pytz.timezone, default is 'UTC'.
        See
        http://en.wikipedia.org/wiki/List_of_tz_database_time_zones
        for a list of valid time zones.
    altitude : float, default 0.
        Altitude from sea level in meters.
    """

    TYPE = "location"

    def __init__(
        self,
        latitude: float,
        longitude: float,
        timezone: str | tz.BaseTzInfo = tz.UTC,
        altitude: Optional[float] = None,
        country: Optional[str] = None,
        state: Optional[str] = None,
    ) -> None:
        self.latitude = latitude
        self.longitude = longitude
        self._timezone = to_timezone(timezone)
        self._altitude = altitude

        # TODO: deduct country, state and timezone from latitude and longitude
        #       with geopy and save to override config file
        self.country = country
        self.state = state

    def __repr__(self):
        attrs = ["latitude", "longitude", "altitude", "timezone"]
        return f"{type(self).__name__}({', '.join(f'{attr}={str(getattr(self, attr))}' for attr in attrs)})"

    def __str__(self):
        attrs = ["latitude", "longitude", "altitude", "timezone"]
        return f"{type(self).__name__}:\n" + "\n\t".join(f"{attr} = {str(getattr(self, attr))}" for attr in attrs)

    @property
    def timezone(self) -> tz.BaseTzInfo:
        return self._timezone

    @property
    def altitude(self) -> float:
        if self._altitude is None:
            return 0.0
        return self._altitude


class LocationException(ResourceError):
    """
    Raise if an error occurred accessing the location.

    """


class LocationUnavailableException(ResourceUnavailableError, LocationException):
    """
    Raise if a configured location access can not be found.

    """



# -*- coding: utf-8 -*-
"""
lories._core._listener
~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from abc import abstractmethod
from typing import Optional, TypeVar

import pandas as pd
from lories._core._entity import _Entity


class _Listener(_Entity):
    @property
    @abstractmethod
    def timestamp(self) -> pd.Timestamp | pd.NaT: ...

    @property
    @abstractmethod
    def runtime(self) -> Optional[float]: ...

    @abstractmethod
    def run(self) -> None: ...

    @abstractmethod
    def locked(self) -> bool: ...

    @abstractmethod
    def has_update(self) -> bool: ...


Listener = TypeVar("Listener", bound=_Listener)



# -*- coding: utf-8 -*-
"""
lories._core._registrator
~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import re
from abc import abstractmethod
from copy import deepcopy
from typing import Any, Collection, Dict, Generic, Optional, Type, TypeVar, Union, overload

from lories._core._configurations import Configurations
from lories._core._configurator import _Configurator
from lories._core._context import _Context
from lories._core._entity import _Entity
from lories.util import validate_key


class _Registrator(_Configurator, _Entity):
    INCLUDES: Collection[str] = ()
    TYPE: str = "registrator"

    @property
    @abstractmethod
    def context(self) -> RegistratorContext: ...

    # noinspection PyShadowingBuiltins
    @classmethod
    def _build_id(
        cls,
        id: Optional[str] = None,
        key: Optional[str] = None,
        context: Optional[RegistratorContext] = None,
        configs: Optional[Configurations] = None,
    ) -> str:
        if configs is not None:
            if id is None:
                if configs.has_member(cls.TYPE) and "id" in configs[cls.TYPE]:
                    id = configs[cls.TYPE]["id"]
                elif "id" in configs:
                    id = configs["id"]
            if key is None:
                if configs.has_member(cls.TYPE) and "key" in configs[cls.TYPE]:
                    key = configs[cls.TYPE]["key"]
                elif "key" in configs:
                    key = configs["key"]
                else:
                    key = validate_key("_".join(re.split(r"[^\w-]", configs.key)))
        if key is None:
            raise ValueError(f"Unable to build '{cls.__name__}' for missing ID")
        if id is None and context is not None and isinstance(context, _Registrator):
            id = f"{context.id}.{key}"
        else:
            id = key
        return id

    @classmethod
    def _build_key(cls, key: Optional[str], configs: Optional[Configurations]) -> str:
        if configs is not None:
            if configs.has_member(cls.TYPE) and "key" in configs[cls.TYPE]:
                key = configs[cls.TYPE]["key"]
            elif "key" in configs:
                key = configs["key"]
            else:
                if configs.has_member(cls.TYPE) and "name" in configs[cls.TYPE]:
                    key = validate_key(configs[cls.TYPE]["name"])
                elif "name" in configs:
                    key = validate_key(configs["name"])
                elif key is None:
                    key = validate_key("_".join(re.split(r"[^\w-]", configs.key)))
        if key is None:
            raise ValueError(f"Unable to build '{cls.__name__}' for missing key")
        return key

    @classmethod
    def _build_name(cls, name: Optional[str], configs: Optional[Configurations]) -> str:
        if configs is not None:
            if configs.has_member(cls.TYPE) and "name" in configs[cls.TYPE]:
                name = configs[cls.TYPE]["name"]
            elif "name" in configs:
                name = configs["name"]
        return name

    @classmethod
    def _build_defaults(
        cls,
        configs: Configurations,
        includes: Optional[Collection[str]] = (),
        **kwargs,
    ) -> Dict[str, Any]:
        defaults = {k: deepcopy(v) for k, v in configs.items() if k in cls.INCLUDES or k in includes}
        defaults.update(**kwargs)
        return defaults


Registrator = TypeVar("Registrator", bound=_Registrator)


# noinspection PyAbstractClass, PyShadowingBuiltins
class _RegistratorContext(_Context[Registrator], Generic[Registrator]):
    TYPE: str = "registrators"

    @overload
    def get_all(self) -> Collection[Registrator]: ...

    @overload
    def get_all(self, type: str, *other: str) -> Collection[Registrator]: ...

    @overload
    def get_all(self, type: Type[Registrator], *other: Type[Registrator]) -> Collection[Registrator]: ...

    def get_all(self, *types: str | Type[Registrator]) -> Collection[Registrator]: ...

    @overload
    def get_first(self) -> Collection[Registrator]: ...

    @overload
    def get_first(self, type: str) -> Collection[Registrator]: ...

    @overload
    def get_first(self, type: str, *other: str) -> Collection[Registrator]: ...

    @overload
    def get_first(self, type: Type[Registrator], *other: Type[Registrator]) -> Collection[Registrator]: ...

    def get_first(self, *types: str | Type[Registrator]) -> Optional[Registrator]: ...

    @overload
    def get_last(self) -> Collection[Registrator]: ...

    @overload
    def get_last(self, type: str, *other: str) -> Collection[Registrator]: ...

    @overload
    def get_last(self, type: Type[Registrator], *other: Type[Registrator]) -> Collection[Registrator]: ...

    def get_last(self, *types: str | Type[Registrator]) -> Optional[Registrator]: ...

    @overload
    def has_type(self, type: str, *other: str) -> bool: ...

    @overload
    def has_type(self, type: Type[Registrator], *other: Type[Registrator]) -> bool: ...

    def has_type(self, *types: str | Type[Registrator]) -> bool: ...


RegistratorContext = TypeVar(
    name="RegistratorContext",
    bound=Union[_RegistratorContext[_Registrator], _Registrator],
)



# -*- coding: utf-8 -*-
"""
lories.application.view.login.authentication
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import base64
import os
import secrets
from typing import Callable, Dict, List, Optional, Union

import flask
from dash import Dash
from dash_auth.auth import Auth
from flask_login import current_user, login_user
from werkzeug.wrappers import Request, Response

import pandas as pd
from lories.application.interface import InterfaceException
from lories.application.view.login.user import User
from lories.core import Configurations, Configurator

RANDOM_STRING_CHARS = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"

UserGroups = Dict[str, List[str]]


class Authentication(Configurator, Auth):
    TYPE: str = "auth"

    __users: Dict[str, User]
    __groups: Optional[Union[UserGroups, Callable[[str], UserGroups]]] = None

    _login_route: str

    def __init__(
        self,
        app: Dash,
        configs: Configurations,
        login_route: str = "auth",
        public_routes: Optional[list] = None,
        secret_key: str = None,
    ):
        if public_routes is None:
            public_routes = []
        if login_route not in public_routes:
            public_routes.append(login_route)
        super().__init__(app=app, public_routes=public_routes)
        self.configure(configs)
        if secret_key is None:
            secret_key = os.getenv("SECRET_KEY", get_random_string(32))
        app.server.secret_key = secret_key

        self._login_route = login_route
        self.__groups = None
        self.__users = {}

        # TODO: Read users from file and add user creation and editing via commandline
        self.add_user("admin", "admin")

    def add_user(self, username: str, password: str) -> None:
        user = User.from_password(username, password)
        self.__users[user.name] = user

    def get_user(self, username: str) -> Optional[User]:
        return self.__users.get(username, None)

    def request_user(self, request: Optional[Request]) -> Optional[User]:
        if request is None:
            request = flask.request
        header = request.headers.get("Authorization", None)
        if not header:
            return None

        username_password = base64.b64decode(header.split("Basic ")[1])
        username_password_utf8 = username_password.decode("utf-8")
        username, password = username_password_utf8.split(":", 1)
        user = self.__users.get(username, None)
        if user is not None:
            user.authenticate(password)
        return user

    def is_authorized(self) -> bool:
        authenticated = current_user.is_authenticated
        if authenticated:
            # TODO: Validate if user is authorized to view route
            try:
                flask.session["user"] = {"email": current_user.id, "groups": []}
                if callable(self.__groups):
                    flask.session["user"]["groups"] = self.__groups(current_user.id)
                elif self.__groups:
                    flask.session["user"]["groups"] = self.__groups.get(current_user.id, [])
            except RuntimeError:
                self._logger.warning("Session is not available. Have you set a secret key?")
        return authenticated

    def login_user(self, username: str, password: str) -> bool:
        user = self.get_user(username)
        if user is not None:
            if user.authenticate(password):
                # TODO: Make remember arguments configurable
                remember = True
                remember_duration = pd.Timedelta(days=7)
                return login_user(user, remember=remember, duration=remember_duration)
            else:
                raise AuthenticationException("Incorrect Password")
        raise AuthenticationException(f"Unknown user '{username}'")

    def login_request(self) -> Response:
        # return flask.redirect(flask.url_for("auth", next_page="/"))
        return self.app.server.make_response(flask.redirect("/login"))


class AuthenticationException(InterfaceException):
    """
    Raise if an error occurred authenticating a session or user.

    """

    reason: str

    def __init__(self, reason: str) -> None:
        super().__init__(reason)
        self.reason = reason


def get_random_string(length, allowed_chars=RANDOM_STRING_CHARS):
    """
    Return a securely generated random string.

    The bit length of the returned value can be calculated with the formula:
        log_2(len(allowed_chars)^length)

    For example, with default `allowed_chars` (26+26+10), this gives:
      * length: 12, bit length =~ 71 bits
      * length: 22, bit length =~ 131 bits
    """
    return "".join(secrets.choice(allowed_chars) for _ in range(length))



# -*- coding: utf-8 -*-
"""
lories.application.view.pages.footer
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations


# TODO Implement footer with "Powered by ISC Konstanz e.V." or similar
class PageFooter:
    pass



# -*- coding: utf-8 -*-
"""
lories._core._constant
~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, Type


# noinspection PyPep8Naming, PyShadowingBuiltins
class _Constant(ABC, str):
    @property
    @abstractmethod
    def type(self) -> Type: ...

    @property
    @abstractmethod
    def key(self) -> str: ...

    @property
    @abstractmethod
    def name(self) -> str: ...

    @property
    @abstractmethod
    def unit(self) -> Optional[str]: ...

    @abstractmethod
    def full_name(self, unit: bool = False) -> str: ...



# -*- coding: utf-8 -*-
"""
lories.components.weather.provider
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides the :class:`lories.components.weather.provider.WeatherProvider`, used as
reference to calculate e.g. photovoltaic installations generated power. The provided
environmental data contains temperatures and horizontal solar irradiation, which can be used,
to calculate the effective irradiance on defined, tilted photovoltaic systems.

"""

from __future__ import annotations

from typing import Optional

from lories.components.weather import Weather, WeatherForecast
from lories.typing import Configurations, ContextArgument


# noinspection SpellCheckingInspection
class WeatherProvider(Weather):
    __forecast: WeatherForecast

    # noinspection PyTypeChecker
    def __init__(
        self,
        context: ContextArgument,
        configs: Optional[Configurations] = None,
        **kwargs,
    ) -> None:
        super().__init__(context=context, configs=configs, **kwargs)
        forecast_configs = configs.get_member(WeatherForecast.TYPE, ensure_exists=True)
        forecast_configs.set("key", "forecast", replace=False)
        self.__forecast = WeatherForecast(self, forecast_configs)

    def _at_configure(self, configs: Configurations) -> None:
        super()._at_configure(configs)
        if self.__forecast.is_enabled():
            self.__forecast.configure(configs.get_member(WeatherForecast.TYPE, ensure_exists=True))

    def _on_configure(self, configs: Configurations) -> None:
        super()._on_configure(configs)
        if self.__forecast.is_enabled() and len(self.__forecast.data) == 0:
            self.__forecast.configs.enabled = False

    def activate(self) -> None:
        super().activate()
        if self.__forecast.is_enabled():
            self.__forecast.activate()

    def deactivate(self) -> None:
        super().deactivate()
        if self.__forecast.is_active():
            self.__forecast.deactivate()

    @property
    def forecast(self) -> WeatherForecast:
        return self.__forecast



# -*- coding: utf-8 -*-
"""
lories.core.constants
~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from abc import ABCMeta
from collections.abc import Callable
from typing import Iterable, Iterator, List, MutableSequence

from lories._core import _Constant  # noqa


class _ConstantsMeta(ABCMeta):
    _instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(_ConstantsMeta, cls).__call__(*args, **kwargs)
        return cls._instance


# noinspection PyShadowingNames
class Constants(MutableSequence[_Constant], metaclass=_ConstantsMeta):
    _constants: List[_Constant]

    def __init__(self, constants=()) -> None:
        self._constants = [*constants]

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join(str(c) for c in self._constants)})"

    def __str__(self) -> str:
        return f"{type(self).__name__}:\n\t" + "\n\t".join(f"{c.key} = {repr(c)}" for c in self._constants)

    def __contains__(self, constant: str | _Constant) -> bool:
        if isinstance(constant, str):
            return any(constant == r.key for r in self._constants)
        return constant in self._constants

    def __getitem__(self, index: Iterable[str] | str | int):
        if isinstance(index, str):
            for constant in self._constants:
                if constant.key == index:
                    return constant
        if isinstance(index, Iterable):
            return type(self)([r for r in self._constants if r.key == index])
        raise KeyError(index)

    def __setitem__(self, index: int, constant: _Constant) -> None:
        self._constants[index] = constant

    def __delitem__(self, index: int) -> None:
        del self._constants[index]

    def __iter__(self) -> Iterator[_Constant]:
        return iter(self._constants)

    def __len__(self) -> int:
        return len(self._constants)

    def __add__(self, other):
        return type(self)([*self, *other])

    def insert(self, index: int, constant: _Constant):
        self._constants.insert(index, constant)

    def extend(self, constants: Iterable[_Constant]) -> None:
        self._constants.extend(constants)

    # noinspection PyShadowingBuiltins
    def filter(self, filter: Callable[[_Constant], bool]):
        return type(self)([constant for constant in self._constants if filter(constant)])



# -*- coding: utf-8 -*-
"""
lories.components.tariff.entsoe
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import pandas as pd
from lories import Constant
from lories.components.tariff import Tariff, TariffProvider, register_tariff_type
from lories.connectors.entsoe import EntsoeConnector
from lories.typing import Configurations


# noinspection SpellCheckingInspection
@register_tariff_type("entsoe", "entso_e")
class EntsoeProvider(TariffProvider):
    PRICE_DAY_AHEAD = Constant(float, "price_day_ahead", name="Day-Ahead Tariff Price", unit="€/MWh")

    _offset: float = 0

    def configure(self, configs: Configurations) -> None:
        super().configure(configs)
        self._offset = configs.get_float("offset", default=0)

        entsoe_connector = EntsoeConnector(
            self,
            key="entsoe_connector",
            name="ENTSO-e Connector",
            configs=configs,
        )
        self.connectors.add(entsoe_connector)
        self.data.add(
            EntsoeProvider.PRICE_DAY_AHEAD,
            method=EntsoeConnector.DAY_AHEAD,
            aggregate="mean",
            connector=entsoe_connector.id,
            logger={"enabled": False},
        )

    def activate(self) -> None:
        super().activate()
        self.data.register(self._on_tariff_received, EntsoeProvider.PRICE_DAY_AHEAD, unique=False)

    def _on_tariff_received(self, data: pd.DataFrame) -> None:
        timestamp = data.index[0]
        import_data = data[EntsoeProvider.PRICE_DAY_AHEAD] / 10.0 + self._offset
        import_channel = self.data.get(Tariff.PRICE_IMPORT)
        import_channel.set(timestamp, import_data)



# -*- coding: utf-8 -*-
"""
lories.connectors.sql.database
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from collections import OrderedDict
from typing import Any, Dict, Iterator, Mapping, Optional

from sqlalchemy import Connection, Dialect, Engine, create_engine, text
from sqlalchemy.exc import SQLAlchemyError

import pandas as pd
import pytz as tz
from lories.connectors import ConnectionError, Database, DatabaseException, register_connector_type
from lories.connectors.sql import Schema, Table
from lories.core.configs import ConfigurationError
from lories.data.util import hash_value
from lories.typing import Configurations, Resources, Timestamp
from lories.util import to_timezone

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


@register_connector_type("sql")
class SqlDatabase(Database, Mapping[str, Table]):
    dialect: Dialect

    host: str
    port: int

    user: str
    password: str
    database: str

    engine: Engine
    _schema: Schema
    _connection: Connection = None

    __tables: Dict[str, Table]

    @property
    def connection(self) -> Connection:
        if not self.is_connected():
            raise ConnectionError(self, "SQL connection not open")
        return self._connection

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.__tables = OrderedDict()

    def __iter__(self) -> Iterator[str]:
        return iter(self.__tables)

    def __len__(self) -> int:
        return len(self.__tables)

    def __contains__(self, table: str | Table) -> bool:
        if isinstance(table, str):
            return table in self.__tables.keys()
        if isinstance(table, Table):
            return table in self.__tables.values()
        return False

    def __getitem__(self, name: str) -> Table:
        return self.__tables[name]

    # noinspection PyShadowingBuiltins, PyProtectedMember
    def _get_vars(self) -> Dict[str, Any]:
        vars = super()._get_vars()
        vars.pop("_schema", None)
        if self.is_configured():
            vars["dialect"] = self.dialect.name
            vars["host"] = self.host
            vars["port"] = self.port
            vars["user"] = self.user
            vars["database"] = self.database
        vars["tables"] = f"[{', '.join(c.name for c in self.__tables.values())}]"
        return vars

    def configure(self, configs: Configurations) -> None:
        super().configure(configs)

        self.host = configs.get("host")
        self.port = configs.get_int("port")

        self.user = configs.get("user")
        self.password = configs.get("password")

        self.database = configs.get("database")

        dialect = configs.get("dialect").lower()
        if dialect == "mysql":
            prefix = "mysql+pymysql://"
        elif dialect == "mariadb":
            prefix = "mariadb+pymysql://"
        elif dialect == "postgresql":
            prefix = "postgresql+psycopg2://"
        else:
            raise ConfigurationError(f"Unsupported database type: {dialect}")
        try:
            self.engine = create_engine(
                url=f"{prefix}{self.user}:{self.password}@{self.host}:{self.port}/{self.database}",
                pool_recycle=-1,
            )
            self.dialect = self.engine.dialect

            self._schema = Schema(self.dialect)
            self._schema.configure(configs.get_member("tables", defaults={}))

        except SQLAlchemyError as e:
            raise ConfigurationError(f"Unable to create database engine: {str(e)}")

    def connect(self, resources: Resources) -> None:
        self._logger.debug(f"Connecting to {self.dialect.name} database {self.database}@{self.host}:{self.port}")
        try:
            self._connection = self.engine.connect()

            # Make sure the connection timezone is UTC
            now = pd.Timestamp.now()
            self._set_timezone(tz.UTC)
            if self._select_timezone().utcoffset(now).seconds != 0:
                raise DatabaseException(self, "Error setting connection timezone to UTC")

            self.__tables = self._schema.connect(self.engine, resources)

        except SQLAlchemyError as e:
            self._raise(e)

    def disconnect(self) -> None:
        if self._connection is not None:
            self._connection.close()
            self._logger.debug("Disconnected from the database")

    def _select_timezone(self) -> tz.BaseTzInfo:
        if self.dialect.name == "postgresql":
            query = "SHOW TIMEZONE"
        elif self.dialect.name in ("mariadb", "mysql"):
            query = "SELECT @@session.time_zone"
        # elif self.dialect.name == 'sqlite':
        #     query = "SELECT datetime('now')"
        else:
            raise NotImplementedError(f"Timezone setting not implemented for dialect: {self.dialect.name}")
        try:
            result = self.connection.execute(text(query))
            timezone = result.scalar()
            return to_timezone(timezone)
        except KeyError:
            raise ValueError(f"Unsupported database type: {self.dialect.name}")
        except SQLAlchemyError as e:
            raise RuntimeError(f"Error fetching timezone: {e}")

    def _set_timezone(self, timezone: tz.BaseTzInfo) -> None:
        # tz_offset = pd.Timestamp.now(timezone).strftime("%:z")
        tz_offset = pd.Timestamp.now(timezone).strftime("%z")
        tz_offset = tz_offset[:3] + ":" + tz_offset[3:]

        if self.dialect.name == "postgresql":
            query = f"SET TIME ZONE '{tz_offset}'"
        elif self.dialect.name in ("mysql", "mariadb"):
            query = f"SET time_zone = '{tz_offset}'"
        else:
            raise NotImplementedError(f"Timezone setting not implemented for dialect: {self.dialect.name}")

        self.connection.execute(text(query))
        self.connection.commit()

    def hash(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        method: Literal["MD5", "SHA1", "SHA256", "SHA512"] = "MD5",
        encoding: str = "UTF-8",
    ) -> Optional[str]:
        hashes = []
        try:
            for table_schema, schema_resources in resources.groupby("schema"):
                for table_name, table_resources in schema_resources.groupby(lambda c: c.get("table", default=c.group)):
                    if table_name not in self.__tables:
                        raise DatabaseException(self, f"Table '{table_name}' not available")

                    table = self.get(table_name)
                    select = table.hash(table_resources, start, end, method=method)
                    result = self.connection.execute(select)

                    # noinspection PyTypeChecker
                    if result.rowcount < 1:
                        continue

                    table_hashes = [r[0] for r in result.fetchall()]
                    if len(table_hashes) > 1:
                        table_hash = hash_value(",".join(table_hashes), method, encoding)
                    else:
                        table_hash = table_hashes[0]
                    hashes.append(table_hash)

        except SQLAlchemyError as e:
            self._raise(e)

        if len(hashes) == 0:
            return None
        elif len(hashes) == 1:
            return hashes[0]

        return hash_value(",".join(hashes), method, encoding)

    def exists(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> bool:
        try:
            for table_schema, schema_resources in resources.groupby("schema"):
                for table_name, table_resources in schema_resources.groupby(lambda c: c.get("table", default=c.group)):
                    if table_name not in self.__tables:
                        raise DatabaseException(self, f"Table '{table_name}' not available")

                    table = self.get(table_name)
                    select = table.exists(table_resources, start, end)
                    result = self.connection.execute(select)

                    # noinspection PyTypeChecker
                    if result.rowcount < 1:
                        continue
                    count = result.scalar()
                    if count is None or int(count) > 1:
                        return True
        except SQLAlchemyError as e:
            self._raise(e)
        return False

    # noinspection PyUnresolvedReferences, PyTypeChecker
    def read(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> pd.DataFrame:
        results = []
        try:
            for table_schema, schema_resources in resources.groupby("schema"):
                for table_name, table_resources in schema_resources.groupby(lambda c: c.get("table", default=c.group)):
                    table_key = table_name if table_schema is None else f"{table_schema}.{table_name}"
                    if table_key not in self.__tables:
                        raise DatabaseException(self, f"Table '{table_key}' not available")

                    table = self.get(table_key)
                    if start is None and end is None:
                        select = table.read(table_resources, order_by="desc").limit(1)
                    else:
                        select = table.read(table_resources, start, end)

                    result = self.connection.execute(select)
                    if result.rowcount > 0:
                        result_data = table.extract(table_resources, result)
                        if not result_data.empty:
                            results.append(result_data)
        except SQLAlchemyError as e:
            self._raise(e)

        if len(results) == 0:
            return pd.DataFrame()
        results = sorted(results, key=lambda d: min(d.index))
        return pd.concat(results, axis="columns")

    # noinspection PyUnresolvedReferences, PyTypeChecker
    def read_first(self, resources: Resources) -> pd.DataFrame:
        results = []
        try:
            for table_schema, schema_resources in resources.groupby("schema"):
                for table_name, table_resources in schema_resources.groupby(lambda c: c.get("table", default=c.group)):
                    if table_name not in self.__tables:
                        raise DatabaseException(self, f"Table '{table_name}' not available")

                    table = self.get(table_name)
                    select = table.read(table_resources, order_by="asc").limit(1)
                    result = self.connection.execute(select)
                    if result.rowcount > 0:
                        result_data = table.extract(table_resources, result)
                        if not result_data.empty:
                            results.append(result_data)
        except SQLAlchemyError as e:
            self._raise(e)

        if len(results) == 0:
            return pd.DataFrame()
        results = sorted(results, key=lambda d: min(d.index))
        return pd.concat(results, axis="columns")

    # noinspection PyUnresolvedReferences, PyTypeChecker
    def read_last(self, resources: Resources) -> pd.DataFrame:
        results = []
        try:
            for table_schema, schema_resources in resources.groupby("schema"):
                for table_name, table_resources in schema_resources.groupby(lambda c: c.get("table", default=c.group)):
                    if table_name not in self.__tables:
                        raise DatabaseException(self, f"Table '{table_name}' not available")

                    table = self.get(table_name)
                    select = table.read(table_resources, order_by="desc").limit(1)
                    result = self.connection.execute(select)
                    if result.rowcount > 0:
                        result_data = table.extract(table_resources, result)
                        if not result_data.empty:
                            results.append(result_data)
        except SQLAlchemyError as e:
            self._raise(e)

        if len(results) == 0:
            return pd.DataFrame()
        results = sorted(results, key=lambda d: min(d.index))
        return pd.concat(results, axis="columns")

    # noinspection PyTypeChecker
    def write(self, data: pd.DataFrame) -> None:
        try:
            for table_schema, schema_resources in self.resources.groupby("schema"):
                for table_name, table_resources in schema_resources.groupby(lambda c: c.get("table", default=c.group)):
                    if table_name not in self.__tables:
                        raise DatabaseException(self, f"Table '{table_name}' not available")
                    table_data = data.loc[:, [r.id for r in table_resources if r.id in data.columns]]
                    table_data = table_data.dropna(axis="index", how="all")
                    if table_data.empty:
                        continue
                    table = self.get(table_name)
                    insert = table.write(table_resources, table_data)
                    self._logger.debug(insert)
                    self.connection.execute(insert)

            self.connection.commit()

        except SQLAlchemyError as e:
            self._raise(e)

    def delete(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> None:
        try:
            for table_schema, schema_resources in resources.groupby("schema"):
                for table_name, table_resources in schema_resources.groupby(lambda c: c.get("table", default=c.group)):
                    if table_name not in self.__tables:
                        raise DatabaseException(self, f"Table '{table_name}' not available")
                    table = self.get(table_name)
                    delete = table.delete(table_resources, start, end)
                    self._logger.debug(delete)
                    self.connection.execute(delete)

            self.connection.commit()

        except SQLAlchemyError as e:
            self._raise(e)

    # noinspection PyProtectedMember
    def is_connected(self) -> bool:
        if self._connection is None:
            return False
        return not self._connection._is_disconnect

    def _raise(self, e: SQLAlchemyError):
        if "syntax" in str(e).lower():
            raise DatabaseException(self, str(e))
        else:
            raise ConnectionError(self, str(e))



# -*- coding: utf-8 -*-
"""
lories.components.tariff.tariff
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from collections.abc import Callable
from typing import Optional, Type, TypeVar

from lories.components import Component, register_component_type
from lories.core import Constant, ResourceError
from lories.core.activator import ActivatorMeta
from lories.core.register import Registry
from lories.typing import Configurations, ContextArgument
from lories.util import validate_key


# noinspection PyShadowingBuiltins
def register_tariff_type(
    type: str,
    *alias: str,
    factory: Callable[[ContextArgument, Optional[Configurations]], TariffType] = None,
    replace: bool = False,
) -> Callable[[Type[TariffType]], Type[TariffType]]:
    # noinspection PyShadowingNames
    def _register(cls: Type[TariffType]) -> Type[TariffType]:
        registry.register(cls, type, *alias, factory=factory, replace=replace)
        return cls

    return _register


class TariffMeta(ActivatorMeta):
    def __call__(cls, context: ContextArgument, configs: Configurations, **kwargs) -> TariffType:
        _type = validate_key(configs.get("type", default="default"))
        _cls = cls._get_class(_type)
        if cls != _cls:
            return _cls(context, configs, **kwargs)

        return super().__call__(context, configs, **kwargs)

    # noinspection PyMethodMayBeStatic, PyShadowingBuiltins
    def _get_class(cls: Type[TariffType], type: str) -> Type[TariffType]:
        if type in ["default", "virtual"]:
            return cls
        elif registry.has_type(type):
            registration = registry.from_type(type)
            return registration.type

        raise ResourceError(f"Unknown tariff type '{type}'")


# noinspection SpellCheckingInspection
@register_component_type("tariff")
class Tariff(Component, metaclass=TariffMeta):
    PRICE_IMPORT = Constant(float, "price_import", name="Import Tariff Price", unit="ct/kWh")
    PRICE_EXPORT = Constant(float, "price_export", name="Export Tariff Price", unit="ct/kWh")


TariffType = TypeVar("TariffType", bound=Tariff)

registry = Registry[Tariff]()



# -*- coding: utf-8 -*-
"""
lories.core.resource
~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
from collections import OrderedDict
from logging import Logger
from typing import Any, Dict, List, Optional, Type

from lories._core._resource import Resource as ResourceType  # noqa
from lories._core._resource import _Resource  # noqa
from lories.core.configs import ConfigurationError
from lories.core.errors import ResourceError
from lories.core.resources import Resources
from lories.util import parse_type, update_recursive, validate_key


class Resource(_Resource):
    __configs: OrderedDict[str, Any]

    _group: str
    _unit: Optional[str]
    _type: Type[Any]

    _logger: Logger

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        id: str = None,
        key: str = None,
        name: Optional[str] = None,
        group: Optional[str] = None,
        unit: Optional[str] = None,
        type: Type | str = None,
        **configs: Any,
    ) -> None:
        super().__init__(id=id, key=key, name=name)
        self._logger = logging.getLogger(self.__module__)

        if group is None:
            group = "_".join(self._id.split(".")[:-1])
        self._group = self._assert_group(group)
        self._unit = self._assert_unit(unit)
        self._type = self._assert_type(parse_type(type))
        self.__configs = OrderedDict(configs)

    @classmethod
    def _assert_group(cls, __group: str) -> str:
        if __group is None:
            raise ResourceError(f"Invalid {cls.__name__}, missing specified 'group'")
        _group = validate_key(__group)
        if _group != __group:
            raise ResourceError(f"Invalid characters in '{cls.__name__}' group: " + __group)
        return _group

    @classmethod
    def _assert_unit(cls, __unit: str) -> str:
        # TODO: Assert unit to be somewhat alphanumeric with certain symbols allowed
        return __unit

    # noinspection PyShadowingBuiltins
    @classmethod
    def _assert_type(cls, __type: Type[Any]) -> Type[Any]:
        # TODO: Validate type to be valid pandas dtypes
        if not isinstance(__type, type):
            raise ResourceError(f"Invalid {cls.__name__}, with type '{__type}' being not allowed.")
        return __type

    def __contains__(self, attr: str) -> bool:
        return attr in self._get_attrs()

    def __getattr__(self, attr: str) -> Any:
        # __getattr__ gets called when the item is not found via __getattribute__
        # To avoid recursion, call __getattribute__ directly to get components dict
        configs = Resource.__getattribute__(self, f"_{Resource.__name__}__configs")
        if attr in configs.keys():
            return configs[attr]
        raise AttributeError(f"'{type(self).__name__}' object has no configuration '{attr}'")

    def __getitem__(self, attr: str) -> Any:
        value = self.get(attr)
        if value is not None:
            return value
        raise KeyError(attr)

    def get(self, attr: str, default: Optional[Any] = None) -> Any:
        return self._get_vars().get(attr, default)

    def _get_attrs(self) -> List[str]:
        return ["id", "key", "name", "group", "type", "unit", *self._copy_configs().keys()]

    def _get_vars(self) -> Dict[str, Any]:
        return OrderedDict(
            id=self.id,
            key=self.key,
            name=self.name,
            group=self.group,
            type=self.type,
            unit=self.unit,
            **self._copy_configs(),
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.id})"

    def __str__(self) -> str:
        return f"{type(self).__name__}:\n\t" + "\n\t".join(f"{k}={v}" for k, v in self._get_vars().items())

    def full_name(self, unit: bool = False) -> str:
        name = self.name
        if unit and self.unit is not None and len(self.unit) > 0:
            name += f" [{self.unit}]"
        return name

    @property
    def group(self) -> str:
        return self._group

    @property
    def unit(self) -> Optional[str]:
        return self._unit

    @property
    def type(self) -> Type:
        return self._type

    # noinspection PyShadowingBuiltins
    def _update(
        self,
        id: Optional[str] = None,
        key: Optional[str] = None,
        name: Optional[str] = None,
        group: Optional[str] = None,
        type: Optional[str | Type] = None,
        unit: Optional[str] = None,
        **configs: Any,
    ) -> None:
        if id is not None and id != self.id:
            raise ConfigurationError(f"Invalid channel update, trying to change ID from '{self.id}' to '{id}'")
        if key is not None and key != self.key:
            raise ConfigurationError(f"Invalid channel update, trying to change Key from '{self.key}' to '{key}'")
        if name is not None:
            self._name = name
        if group is not None:
            self._group = self._assert_group(group)
        if type is not None:
            self._type = self._assert_type(parse_type(type))
        self._unit = self._assert_unit(unit)
        self.__update_configs(configs)

    def __update_configs(self, configs: Dict[str, Any]) -> None:
        update_recursive(self.__configs, configs)

    def _copy_configs(self) -> Dict[str, Any]:
        return OrderedDict(**self.__configs)

    def _copy_args(self) -> Dict[str, Any]:
        return {
            "id": self._id,
            **self.to_configs(),
        }

    def copy(self) -> ResourceType:
        return self.duplicate()

    def duplicate(self, **changes) -> ResourceType:
        arguments = self._copy_args()
        arguments.update(changes)
        return super().duplicate(**arguments)

    def to_list(self) -> Resources:
        return Resources([self])

    def to_configs(self) -> Dict[str, Any]:
        return {
            "key": self._key,
            "name": self._name,
            "group": self._group,
            "unit": self._unit,
            "type": self._type,
            **self._copy_configs(),
        }
